#if (defined(_WIN32) || defined(_WIN64)) && !defined(_CRT_SECURE_NO_WARNINGS)
	// This disables annoying warnings VC++ gives for use of C standard library.  Shut the mouth up.
	// This needs to come before header inclusion.
	#define _CRT_SECURE_NO_WARNINGS
#endif

#include <ysclass.h>

#include "ysdnm.h"


// tanFov : (X or Y)/Z
//       Z
//       |      /
//       |     /
//       |    /
//       |FOV/
//       |  /
//       | /
//       |/
//       +-------X
YSBOOL YsDnm::IsPointVisible
    (int np,const YsVec3 p[],
     const YsMatrix4x4 &viewModelTfm,
	 const double &nearZ,const double &farZ,const double &tanFov)
{
	YsVec3 corner;
	int i,nFar,nNear;
	int nRight,nRightRear,nLeft,nLeftRear,nTop,nTopRear,nBottom,nBottomRear;

	// Check
	//   Right threshold  x=z*tanFov   z*tanFov<x is out
	//   Left  threshold  x=-z*tanFov  x<-z*tanFov is out
	//   Top threshold    y=z*tanFov   z*tanFov<y is out
	//   Bottom Threshold y=-z*tanFov  y<-z*tanFov is out
	nFar=0;
	nNear=0;
	nRight=0;
	nRightRear=0;
	nLeft=0;
	nLeftRear=0;
	nTop=0;
	nTopRear=0;
	nBottom=0;
	nBottomRear=0;
	for(i=0; i<np; i++)
	{
		corner=viewModelTfm*p[i];

		YSBOOL inc;
		inc=YSFALSE;
		if(corner.z()<nearZ)
		{
			nNear++;
			inc=YSTRUE;
		}
		if(corner.z()>farZ)
		{
			nFar++;
			inc=YSTRUE;
		}
		if(farZ*tanFov<corner.x())
		{
			nRight++;
			inc=YSTRUE;
		}
		if(corner.z()*tanFov<corner.x())
		{
			nRightRear++;
			inc=YSTRUE;
		}
		if(corner.x()<-farZ*tanFov)
		{
			nLeft++;
			inc=YSTRUE;
		}
		if(corner.x()<-corner.z()*tanFov)
		{
			nLeftRear++;
			inc=YSTRUE;
		}
		if(farZ*tanFov<corner.y())
		{
			nTop++;
			inc=YSTRUE;
		}
		if(corner.z()*tanFov<corner.y())
		{
			nTopRear++;
			inc=YSTRUE;
		}
		if(corner.y()<-farZ*tanFov)
		{
			nBottom++;
			inc=YSTRUE;
		}
		if(corner.y()<-corner.z()*tanFov)
		{
			nBottomRear++;
			inc=YSTRUE;
		}

		if(inc!=YSTRUE)
		{
			break;
		}
	}

	if(nFar==np || nNear==np ||
	   nRight==np || nLeft==np || nTop==np || nBottom==np ||
	   nRightRear==np || nLeftRear==np || nTopRear==np || nBottomRear==np)
	{
		return YSFALSE;
	}

	return YSTRUE;
}

YSBOOL YsDnm::IsBbxVisible
	    (const YsVec3 bbx[2],
	     const YsMatrix4x4 &viewModelTfm,
	     const double &nearZ,const double &farZ,const double &tanFov)
{
	int i,nFar,nNear;
	YsVec3 corner[8],minmax[2],cen;

	corner[0].Set(bbx[0].x(),bbx[0].y(),bbx[0].z());
	corner[1].Set(bbx[0].x(),bbx[0].y(),bbx[1].z());
	corner[2].Set(bbx[0].x(),bbx[1].y(),bbx[0].z());
	corner[3].Set(bbx[0].x(),bbx[1].y(),bbx[1].z());
	corner[4].Set(bbx[1].x(),bbx[0].y(),bbx[0].z());
	corner[5].Set(bbx[1].x(),bbx[0].y(),bbx[1].z());
	corner[6].Set(bbx[1].x(),bbx[1].y(),bbx[0].z());
	corner[7].Set(bbx[1].x(),bbx[1].y(),bbx[1].z());

	corner[0]=viewModelTfm*corner[0];
	corner[1]=viewModelTfm*corner[1];
	corner[2]=viewModelTfm*corner[2];
	corner[3]=viewModelTfm*corner[3];
	corner[4]=viewModelTfm*corner[4];
	corner[5]=viewModelTfm*corner[5];
	corner[6]=viewModelTfm*corner[6];
	corner[7]=viewModelTfm*corner[7];

	// Check
	//   Right threshold  x=z*tanFov   z*tanFov<x is out
	//   Left  threshold  x=-z*tanFov  x<-z*tanFov is out
	//   Top threshold    y=z*tanFov   z*tanFov<y is out
	//   Bottom Threshold y=-z*tanFov  y<-z*tanFov is out
	int nRight,nRightRear,nLeft,nLeftRear,nTop,nTopRear,nBottom,nBottomRear;
	nFar=0;
	nNear=0;
	nRight=0;
	nRightRear=0;
	nLeft=0;
	nLeftRear=0;
	nTop=0;
	nTopRear=0;
	nBottom=0;
	nBottomRear=0;
	for(i=0; i<8; i++)
	{
		YSBOOL inc;
		inc=YSFALSE;
		if(corner[i].z()<nearZ)
		{
			nNear++;
			inc=YSTRUE;
		}
		if(corner[i].z()>farZ)
		{
			nFar++;
			inc=YSTRUE;
		}
		if(farZ*tanFov<corner[i].x())
		{
			nRight++;
			inc=YSTRUE;
		}
		if(corner[i].z()*tanFov<corner[i].x())
		{
			nRightRear++;
			inc=YSTRUE;
		}
		if(corner[i].x()<-farZ*tanFov)
		{
			nLeft++;
			inc=YSTRUE;
		}
		if(corner[i].x()<-corner[i].z()*tanFov)
		{
			nLeftRear++;
			inc=YSTRUE;
		}
		if(farZ*tanFov<corner[i].y())
		{
			nTop++;
			inc=YSTRUE;
		}
		if(corner[i].z()*tanFov<corner[i].y())
		{
			nTopRear++;
			inc=YSTRUE;
		}
		if(corner[i].y()<-farZ*tanFov)
		{
			nBottom++;
			inc=YSTRUE;
		}
		if(corner[i].y()<-corner[i].z()*tanFov)
		{
			nBottomRear++;
			inc=YSTRUE;
		}

		if(inc!=YSTRUE)
		{
			break;
		}
	}

	if(nFar==8 || nNear==8 ||
	   nRight==8 || nLeft==8 || nTop==8 || nBottom==8 ||
	   nRightRear==8 || nLeftRear==8 || nTopRear==8 || nBottomRear==8)
	{
		return YSFALSE;
	}

	return YSTRUE;


//	YsVec3 corner[8];
//
//	corner[0].Set(bbx[0].x(),bbx[0].y(),bbx[0].z());
//	corner[1].Set(bbx[0].x(),bbx[0].y(),bbx[1].z());
//	corner[2].Set(bbx[0].x(),bbx[1].y(),bbx[0].z());
//	corner[3].Set(bbx[0].x(),bbx[1].y(),bbx[1].z());
//	corner[4].Set(bbx[1].x(),bbx[0].y(),bbx[0].z());
//	corner[5].Set(bbx[1].x(),bbx[0].y(),bbx[1].z());
//	corner[6].Set(bbx[1].x(),bbx[1].y(),bbx[0].z());
//	corner[7].Set(bbx[1].x(),bbx[1].y(),bbx[1].z());
//
//	return IsPointVisible(8,corner,viewModelTfm,nearZ,farZ,tanFov);
}

int YsShellVisual::globalControl_lightPointSize=3;
unsigned YsShellVisual::globalControl_transparency=0;  // 0:No transparency  255:Use transparency
YSBOOL YsShellVisual::globalControl_useBiSubTree=YSTRUE;
YSBOOL YsShellVisual::globalControl_useDisplayList=YSTRUE;
YSBOOL YsShellVisual::globalControl_triangulateAll=YSFALSE;
YSBOOL YsShellVisual::globalControl_antiAliasing=YSFALSE;

YsShellVisual::YsShellVisual()
{
	flags=0;
	noShadPlg=NULL;
	roundVtx=NULL;
	plgAhpla=NULL;
	invisiNonOpenGL=NULL;
	plgLightFlag=NULL;
	openGlListId=0;
	extBufLink=NULL;
	AllocateD3dExternalVertexBufferLink();
}

YsShellVisual::~YsShellVisual()
{
	if(noShadPlg!=NULL)
	{
		delete noShadPlg;
		noShadPlg=NULL;
	}
	if(roundVtx!=NULL)
	{
		delete roundVtx;
		roundVtx=NULL;
	}
	if(plgAhpla!=NULL)
	{
		delete plgAhpla;
		plgAhpla=NULL;
	}
	if(invisiNonOpenGL!=NULL)
	{
		delete invisiNonOpenGL;
		invisiNonOpenGL=NULL;
	}
	if(plgLightFlag!=NULL)
	{
		delete plgLightFlag;
		plgLightFlag=NULL;
	}
	FreeD3dExternalVertexBufferLink();
}

const YsShellVisual &YsShellVisual::operator=(const YsShellVisual &from)
{
	YsShell *thisShell;
	const YsShell *fromShell;

	thisShell=this;
	fromShell=&from;

	(*thisShell)=(*fromShell);
	plgCenter=from.plgCenter;

	ReleaseD3dExternalVertexBuffer();

	return *this;
}

void YsShellVisual::ReadSrfPostProcess(void)
{
	int i;
	YsShellVertexHandle vtHd;
	YsShellPolygonHandle plHd;
	YsArray <YsVec3> plVtPos;
	YsArray <YsShellPolygonHandle> plHdList;

	vtHd=NULL;
	while((vtHd=FindNextVertex(vtHd))!=NULL)
	{
		GetVertex(vtHd)->freeAttribute0=0;
		GetVertex(vtHd)->SetNormal(YsOrigin());
	}
	plHd=NULL;
	while((plHd=FindNextPolygon(plHd))!=NULL)
	{
		GetPolygon(plHd)->freeAttribute0=0;
		plHdList.Append(plHd);
	}


	for(i=0; i<noShadPlg->GetN(); i++)
	{
		GetPolygon((*noShadPlg)[i])->freeAttribute0|=PLG_NOSHADING;
	}
	for(i=0; i<roundVtx->GetN(); i++)
	{
		GetVertex((*roundVtx)[i])->freeAttribute0|=VTX_ROUND;
	}
	for(i=0; i<plgAhpla->GetN()-1; i+=2)
	{
		int plId;
		plId=(*plgAhpla)[i];
		if(0<=plId && plId<plHdList.GetN())
		{
			plHd=plHdList[plId];
			GetPolygon(plHd)->freeAttribute0&=~0xff00;
			GetPolygon(plHd)->freeAttribute0|=((*plgAhpla)[i+1]&255)<<8;
		}
	}
	for(i=0; i<invisiNonOpenGL->GetN(); i++)
	{
		int plId;
		plId=(*invisiNonOpenGL)[i];
		if(0<=plId && plId<plHdList.GetN())
		{
			plHd=plHdList[plId];
			GetPolygon(plHd)->freeAttribute0|=PLG_INVISINONOPENGL;
		}
	}

	light.Set(0,NULL);
	for(i=0; i<plgLightFlag->GetN(); i++)
	{
		int plId;
		plId=(*plgLightFlag)[i];
		if(0<=plId && plId<plHdList.GetN())
		{
			plHd=plHdList[plId];
			GetPolygon(plHd)->freeAttribute0|=PLG_LIGHT;

			light.Increment();

			light.GetEnd().plHd=plHd;
			GetCenterOfPolygon(light.GetEnd().cen,plHd);
			GetNormalOfPolygon(light.GetEnd().nom,plHd);

			int nPlVt;
			const YsShellVertexHandle *plVtHd;
			GetVertexListOfPolygon(nPlVt,plVtHd,plHd);

			int j;
			light.GetEnd().rad=0.0;
			for(j=0; j<nPlVt; j++)
			{
				YsVec3 pos;
				GetVertexPosition(pos,plVtHd[j]);
				light.GetEnd().rad=YsGreater(light.GetEnd().rad,(pos-light.GetEnd().cen).GetLength());
			}
			GetColorOfPolygon(light.GetEnd().col,plHd);
		}
	}

	plHd=NULL;
	while((plHd=FindNextPolygon(plHd))!=NULL)
	{
		if(GetVertexListOfPolygon(plVtPos,plHd)==YSOK)
		{
	   		YsVec3 nom;
			GetNormalOfPolygon(nom,plHd);
			if(nom==YsOrigin())
			{
				GetPolygon(plHd)->freeAttribute0|=PLG_TWOSIDED;
				if(YsGetAverageNormalVector(nom,plVtPos.GetN(),plVtPos)==YSOK)
				{
					SetNormalOfPolygon(plHd,nom);
				}
			}

			if(nom.IsNormalized()!=YSTRUE)
			{
				nom.Normalize();
				SetNormalOfPolygon(plHd,nom);
			}

			if(YsCheckConvexByAngle3(plVtPos.GetN(),plVtPos)!=YSTRUE)
			{
				GetPolygon(plHd)->freeAttribute0|=PLG_CONCAVE;
			}
		}
	}
}

void YsShellVisual::CleanUp(void)
{
	YsShell::CleanUp();
	ReleaseD3dExternalVertexBuffer();
	plgCenter.Set(0,NULL);
	light.Set(0,NULL);
}

YSRESULT YsShellVisual::BeginReadSrf(void)
{
	if(YsShell::BeginReadSrf()==YSOK)
	{
		ReleaseD3dExternalVertexBuffer();

		if(noShadPlg!=NULL)
		{
			delete noShadPlg;
		}
		noShadPlg=new YsArray <YsShellPolygonHandle>;

		if(roundVtx!=NULL)
		{
			delete roundVtx;
		}
		roundVtx=new YsArray <YsShellVertexHandle>;

		if(plgAhpla!=NULL)
		{
			delete plgAhpla;
		}
		plgAhpla=new YsArray <unsigned>;

		if(invisiNonOpenGL!=NULL)
		{
			delete invisiNonOpenGL;
		}
		invisiNonOpenGL=new YsArray <unsigned>;

		if(plgLightFlag!=NULL)
		{
			delete plgLightFlag;
		}
		plgLightFlag=new YsArray <unsigned>;

		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsShellVisual::ReadSrfOneLine(const char str[])
{
	if(YsShell::ReadSrfOneLine(str,noShadPlg,roundVtx)!=YSOK)
	{
		if(strncmp(str,"ZA",2)==0 || strncmp(str,"za",2)==0)
		{
			int ac;
			char *av[256],buf[1024];
			strncpy(buf,str,1023);
			buf[1023]=0;
			if(YsArguments(&ac,av,255,buf)==YSOK)
			{
				int i;
				for(i=1; i<ac-1; i+=2)
				{
					plgAhpla->Append(atoi(av[i]));
					plgAhpla->Append(atoi(av[i+1]));
				}
			}
			return YSOK;
		}
		else if(strncmp(str,"ZL",2)==0 || strncmp(str,"zl",2)==0)
		{
			int ac;
			char *av[256],buf[1024];
			strncpy(buf,str,1023);
			buf[1023]=0;
			if(YsArguments(&ac,av,255,buf)==YSOK)
			{
				int i;
				for(i=1; i<ac; i++)
				{
					plgLightFlag->Append(atoi(av[i]));
				}
			}
			return YSOK;
		}
		else if(strncmp(str,"ZZ",2)==0 || strncmp(str,"zz",2)==0)
		{
			int ac;
			char *av[256],buf[1024];
			strncpy(buf,str,1023);
			buf[1023]=0;
			if(YsArguments(&ac,av,255,buf)==YSOK)
			{
				int i;
				for(i=1; i<ac; i++)
				{
					invisiNonOpenGL->Append(atoi(av[i]));
				}
			}
			return YSOK;
		}
	}
	return YSOK;  // Ignore errors
}

YSRESULT YsShellVisual::EndReadSrf(void)
{
	if(YsShell::EndReadSrf()==YSOK)
	{
		ReadSrfPostProcess();
		RecomputeNormalAndFlipDirection();
		if(globalControl_triangulateAll==YSTRUE)
		{
			YsPrintf("Tessellating...\n");
			TriangulateAll();
		}
		SortPolygonByColor();
		PrecomputePolygCenterAfterSortingByColor();
		delete noShadPlg;
		noShadPlg=NULL;
		delete roundVtx;
		roundVtx=NULL;
		delete plgAhpla;
		plgAhpla=NULL;
		delete invisiNonOpenGL;
		invisiNonOpenGL=NULL;
		delete plgLightFlag;
		plgLightFlag=NULL;

		Encache();

		return YSOK;
	}
	else
	{
		if(noShadPlg!=NULL)
		{
			delete noShadPlg;
			noShadPlg=NULL;
		}
		if(roundVtx!=NULL)
		{
			delete roundVtx;
			roundVtx=NULL;
		}
		if(plgAhpla!=NULL)
		{
			delete plgAhpla;
			plgAhpla=NULL;
		}
		if(invisiNonOpenGL!=NULL)
		{
			delete invisiNonOpenGL;
			invisiNonOpenGL=NULL;
		}

		return YSERR;
	}
}


YSRESULT YsShellVisual::SaveSrf(const char fn[]) const
{
	FILE *fp;
	fp=fopen(fn,"w");
	if(NULL!=fp)
	{
		YsTextFileOutputStream textOut(fp);
		YSRESULT res=SaveSrf(textOut);
		fclose(fp);
		return res;
	}
	return YSERR;
}

YSRESULT YsShellVisual::SaveSrf(YsTextOutputStream &textOut) const
{
	YsShellPolygonHandle plHd;
	YsArray <YsShellPolygonHandle> noShadingPlg;
	YsShellVertexHandle vtHd;
	YsArray <YsShellVertexHandle> roundVtx;

	vtHd=NULL;
	while(NULL!=(vtHd=FindNextVertex(vtHd)))
	{
		if(GetVertex(vtHd)->freeAttribute0&VTX_ROUND)
		{
			roundVtx.Append(vtHd);
		}
	}

	plHd=NULL;
	while(NULL!=(plHd=FindNextPolygon(plHd)))
	{
		if(GetPolygon(plHd)->freeAttribute0&PLG_NOSHADING)
		{
			noShadingPlg.Append(plHd);
		}

	}

	if(YsShell::SaveSrf(textOut,noShadingPlg.GetN(),noShadingPlg,roundVtx.GetN(),roundVtx)==YSOK)
	{
		int plId;
		plHd=NULL;
		plId=0;
		while(NULL!=(plHd=FindNextPolygon(plHd)))
		{
			if(GetPolygon(plHd)->freeAttribute0&PLG_INVISINONOPENGL)
			{
				textOut.Printf("ZZ %d\n",plId);
			}

			int ahpla;
			ahpla=(GetPolygon(plHd)->freeAttribute0>>8)&255;
			if(ahpla!=0)
			{
				textOut.Printf("ZA %d %d\n",plId,ahpla);
			}

			plId++;
		}

		return YSOK;
	}
	return YSERR;
}

void YsShellVisual::CheckLeak(void)
{
	// GCC gives error.
	// printf("%08x %08x %08x %08x\n",(int)noShadPlg,(int)roundVtx,(int)plgAhpla,(int)invisiNonOpenGL);
}

void YsShellVisual::RecomputeNormalAndFlipDirection(void)
{
	YsArray <YsVec3> plVtPos;
	YsShellPolygonHandle plHd;
	int nNoInvert,nInvert;

	nNoInvert=0;
	nInvert=0;
	plHd=NULL;
	while((plHd=FindNextPolygon(plHd))!=NULL)
	{
		if(GetVertexListOfPolygon(plVtPos,plHd)==YSOK)
		{
	   		YsVec3 nom;
			GetNormalOfPolygon(nom,plHd);
			if(nom!=YsOrigin())
			{
				// YSCOORDSYSMODEL prevCoordSys;  Commented out 2006/04/27
				// prevCoordSys=YsCoordSysModel;  Commented out 2006/04/27
				// YsCoordSysModel=YSBLUEIMPULSE; Commented out 2006/04/27
				if(YsCheckFlipDirection3(plVtPos.GetN(),plVtPos,nom)==YSFLIPCCW)
				{
					InvertPolygon(plHd);
					SetNormalOfPolygon(plHd,nom);  // Because InvertPolygon also inverts normal.
					                               // Now I'm fixing the flip direction.
					nInvert++;
				}
				else
				{
					nNoInvert++;
				}
				// YsCoordSysModel=prevCoordSys; Commented out 2006/04/27
			}
		}
	}
	// YsPrintf("# Inverted=%d  # Not Inverted=%d\n",nInvert,nNoInvert);

	// Compute vertex normal
	plHd=NULL;
	while((plHd=FindNextPolygon(plHd))!=NULL)
	{
		if((GetPolygon(plHd)->freeAttribute0&(PLG_TWOSIDED|PLG_NOSHADING))==0)
		{
			YsVec3 nom,vtxNom;
			int i,nPlVt;
			const YsShellVertexHandle *plVtHd;
			nPlVt=GetNumVertexOfPolygon(plHd);
			plVtHd=GetVertexListOfPolygon(plHd);
			GetNormalOfPolygon(nom,plHd);
			for(i=0; i<nPlVt; i++)
			{
				vtxNom=GetVertex(plVtHd[i])->GetNormal();
				vtxNom+=nom;
				GetVertex(plVtHd[i])->SetNormal(vtxNom);
			}
		}
	}

	YsShellVertexHandle vtHd;
	vtHd=NULL;
	while((vtHd=FindNextVertex(vtHd))!=NULL)
	{
		if(GetVertex(vtHd)->freeAttribute0&VTX_ROUND)
		{
			YsVec3 vtxNom;
			vtxNom=GetVertex(vtHd)->GetNormal();
			if(vtxNom.Normalize()==YSOK)
			{
				GetVertex(vtHd)->SetNormal(vtxNom);
			}
			else
			{
				GetVertex(vtHd)->freeAttribute0&=~VTX_ROUND;
			}
		}
	}
}

void YsShellVisual::TriangulateAll(void)
{
	YsArray <YsShellPolygonHandle> newPlHd;
	YsShellPolygonHandle plHd;
	RewindPlgPtr();
	while(NULL!=(plHd=StepPlgPtr()))
	{
		const YsShellPolygon *plg;
		int nPlVt;
		const YsShellVertexHandle *plVtHd;

		GetVertexListOfPolygon(nPlVt,plVtHd,plHd);
		if(nPlVt>3)
		{
			YsVec3 nom;
			YsColor col;
			unsigned int freeAttribute0,freeAttribute1,freeAttribute2;

			plg=GetPolygon(plHd);
			freeAttribute0=plg->freeAttribute0;
			freeAttribute1=plg->freeAttribute1;
			freeAttribute2=plg->freeAttribute2;
			GetNormalOfPolygon(nom,plHd);
			GetColorOfPolygon(col,plHd);
			if((plg->freeAttribute0&PLG_CONCAVE)==0)
			{
				SimpleTriangulation(plHd);
			}
			else if(Triangulate(newPlHd,plHd,YSTRUE)==YSOK)
			{
			}
			else
			{
				SimpleTriangulation(plHd);
			}
		}
	}
}

void YsShellVisual::SimpleTriangulation(YsShellPolygonHandle plHd)
{
	int i;
	YsShellPolygonHandle newPlHd;
	YsShellVertexHandle triVtHd[3];

	YsVec3 nom;
	YsColor col;
	unsigned int freeAttribute0,freeAttribute1,freeAttribute2;

	const YsShellPolygon *plg;
	int nPlVt;
	const YsShellVertexHandle *plVtHd;

	GetVertexListOfPolygon(nPlVt,plVtHd,plHd);

	plg=GetPolygon(plHd);
	freeAttribute0=plg->freeAttribute0;
	freeAttribute1=plg->freeAttribute1;
	freeAttribute2=plg->freeAttribute2;
	GetNormalOfPolygon(nom,plHd);
	GetColorOfPolygon(col,plHd);

	triVtHd[0]=plVtHd[0];
	for(i=2; i<nPlVt-1; i++)
	{
		triVtHd[1]=plVtHd[i];
		triVtHd[2]=plVtHd[i+1];
		newPlHd=AddPolygonH(3,triVtHd);
		GetPolygon(newPlHd)->freeAttribute0=freeAttribute0;
		GetPolygon(newPlHd)->freeAttribute1=freeAttribute1;
		GetPolygon(newPlHd)->freeAttribute2=freeAttribute2;
		SetNormalOfPolygon(newPlHd,nom);
		SetColorOfPolygon(newPlHd,col);
	}

	triVtHd[1]=plVtHd[1];
	triVtHd[2]=plVtHd[2];
	ModifyPolygon(plHd,3,triVtHd);
}

void YsShellVisual::SortPolygonByColor(void)
{
	YsArray <unsigned,4096> plgShadColorList,plgNoShadColorList;
	YsArray <YsShellPolygonHandle,4096> plgShadPlHdList,plgNoShadPlHdList;

	YsArray <unsigned,4096> triShadColorList,triNoShadColorList;
	YsArray <YsShellPolygonHandle,4096> triShadPlHdList,triNoShadPlHdList;

	YsArray <unsigned,4096> quadShadColorList,quadNoShadColorList;
	YsArray <YsShellPolygonHandle,4096> quadShadPlHdList,quadNoShadPlHdList;


	YsShellPolygonHandle plHd;
	YsColor col;
	const YsShellPolygon *plg;

	plHd=NULL;
	while((plHd=FindNextPolygon(plHd))!=NULL)
	{
		plg=GetPolygon(plHd);
		GetColorOfPolygon(col,plHd);

		unsigned colCode,ahpla;
		colCode=col.Bi()+col.Gi()*256+col.Ri()*65536;

		ahpla=(plg->freeAttribute0>>8)&255;
		colCode+=(ahpla<<24);


		switch(GetNumVertexOfPolygon(plHd))
		{
		default:
			if(plg->freeAttribute0&PLG_NOSHADING)
			{
				plgNoShadColorList.Append(colCode);
				plgNoShadPlHdList.Append(plHd);
			}
			else
			{
				plgShadColorList.Append(colCode);
				plgShadPlHdList.Append(plHd);
			}
			break;
		case 3:
			if(plg->freeAttribute0&PLG_NOSHADING)
			{
				triNoShadColorList.Append(colCode);
				triNoShadPlHdList.Append(plHd);
			}
			else
			{
				triShadColorList.Append(colCode);
				triShadPlHdList.Append(plHd);
			}
			break;
		case 4:
			if(plg->freeAttribute0&PLG_NOSHADING)
			{
				quadNoShadColorList.Append(colCode);
				quadNoShadPlHdList.Append(plHd);
			}
			else
			{
				quadShadColorList.Append(colCode);
				quadShadPlHdList.Append(plHd);
			}
			break;
		}
	}

	YsQuickSort <unsigned,YsShellPolygonHandle> (plgShadColorList.GetN(),plgShadColorList,plgShadPlHdList);
	YsQuickSort <unsigned,YsShellPolygonHandle> (plgNoShadColorList.GetN(),plgNoShadColorList,plgNoShadPlHdList);

	YsQuickSort <unsigned,YsShellPolygonHandle> (triShadColorList.GetN(),triShadColorList,triShadPlHdList);
	YsQuickSort <unsigned,YsShellPolygonHandle> (triNoShadColorList.GetN(),triNoShadColorList,triNoShadPlHdList);

	YsQuickSort <unsigned,YsShellPolygonHandle> (quadShadColorList.GetN(),quadShadColorList,quadShadPlHdList);
	YsQuickSort <unsigned,YsShellPolygonHandle> (quadNoShadColorList.GetN(),quadNoShadColorList,quadNoShadPlHdList);

	int i;
	for(i=0; i<triShadPlHdList.GetN(); i++)
	{
		YsShell::plg.MoveItemToEnd((YsListItem <YsShellPolygon> *)triShadPlHdList[i]);
	}
	for(i=0; i<quadShadPlHdList.GetN(); i++)
	{
		YsShell::plg.MoveItemToEnd((YsListItem <YsShellPolygon> *)quadShadPlHdList[i]);
	}
	for(i=0; i<plgShadPlHdList.GetN(); i++)
	{
		YsShell::plg.MoveItemToEnd((YsListItem <YsShellPolygon> *)plgShadPlHdList[i]);
	}

	for(i=0; i<triNoShadPlHdList.GetN(); i++)
	{
		YsShell::plg.MoveItemToEnd((YsListItem <YsShellPolygon> *)triNoShadPlHdList[i]);
	}
	for(i=0; i<quadNoShadPlHdList.GetN(); i++)
	{
		YsShell::plg.MoveItemToEnd((YsListItem <YsShellPolygon> *)quadNoShadPlHdList[i]);
	}
	for(i=0; i<plgNoShadPlHdList.GetN(); i++)
	{
		YsShell::plg.MoveItemToEnd((YsListItem <YsShellPolygon> *)plgNoShadPlHdList[i]);
	}
}

void YsShellVisual::PrecomputePolygCenterAfterSortingByColor(void)
{
	int i;
	YsShellPolygonHandle plHd;
	YsVec3 cen;
	plgCenter.Set(GetNumPolygon(),NULL);
	plHd=NULL;
	i=0;
	while((plHd=FindNextPolygon(plHd))!=NULL)
	{
		GetCenterOfPolygon(cen,plHd);
		plgCenter[i]=cen;
		i++;
	}
}

YSRESULT YsShellVisual::LoadSrf(const char fn[])
{
	YSRESULT res=YSERR;
	FILE *fp=fopen(fn,"r");
	if(NULL!=fp)
	{
		res=LoadSrf(fp);
		fclose(fp);
	}
	return res;
}

YSRESULT YsShellVisual::LoadSrf(FILE *fp)
{
	YsTextFileInputStream inStream(fp);
	return LoadSrf(inStream);
}

YSRESULT YsShellVisual::LoadSrf(YsTextInputStream &inStream)
{
	YsString buf;
	if(BeginReadSrf()==YSOK)
	{
		while(NULL!=inStream.Gets(buf))
		{
			if(ReadSrfOneLine(buf)!=YSOK)
			{
				EndReadSrf();
				goto ERREND;
			}
		}
		if(EndReadSrf()==YSOK)
		{
			return YSOK;
		}
	}

ERREND:

	if(noShadPlg!=NULL)
	{
		delete noShadPlg;
		noShadPlg=NULL;
	}
	if(roundVtx!=NULL)
	{
		delete roundVtx;
		roundVtx=NULL;
	}
	if(plgAhpla!=NULL)
	{
		delete plgAhpla;
		plgAhpla=NULL;
	}
	if(invisiNonOpenGL!=NULL)
	{
		delete invisiNonOpenGL;
		invisiNonOpenGL=NULL;
	}
	return YSERR;
}

static YSRESULT YsMakePlaneProjectionMatrix(YsMatrix4x4 &tfm,const YsVec3 &o,const YsVec3 &n,const YsVec3 &q)  // q:Direction of projection
{
	const double qn=q*n;
	const double on=o*n;

	if(qn!=0.0)
	{
		tfm.Set(1,1,1.0-(n.x()*q.x())/qn);
		tfm.Set(1,2,-n.y()*q.x()/qn);
		tfm.Set(1,3,-n.z()*q.x()/qn);
		tfm.Set(1,4,q.x()*on/qn);

		tfm.Set(2,1,-n.x()*q.y()/qn);
		tfm.Set(2,2,1.0-n.y()*q.y()/qn);
		tfm.Set(2,3,-n.z()*q.y()/qn);
		tfm.Set(2,4,q.y()*on/qn);

		tfm.Set(3,1,-n.x()*q.z()/qn);
		tfm.Set(3,2,-n.y()*q.z()/qn);
		tfm.Set(3,3,1.0-n.z()*q.z()/qn);
		tfm.Set(3,4,q.z()*on/qn);

		tfm.Set(4,1,0.0);
		tfm.Set(4,2,0.0);
		tfm.Set(4,3,0.0);
		tfm.Set(4,4,1.0);

		return YSOK;
	}

	return YSERR;
}

void YsShellVisual::DrawShadowVisual
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir,
	     const double &nearZ,const double &farZ,const double &tanFov) const
{
	YsMatrix4x4 prjMat(YSFALSE);
	YsMakePlaneProjectionMatrix(prjMat,terOrg,terNom,lightDir);


	YsMatrix4x4 viewShadowModelTfm(YSFALSE);;
	viewShadowModelTfm=viewTfm;
	viewShadowModelTfm*=prjMat;
	// viewShadowModelTfm.Translate(0.0,elv,0.0);
	// viewShadowModelTfm.Scale(1.0,0.0,1.0);
	viewShadowModelTfm.Translate(pos);
	viewShadowModelTfm.RotateXZ(att.h());
	viewShadowModelTfm.RotateZY(att.p());
	viewShadowModelTfm.RotateXY(att.b());

	YsVec3 bbx[2],corner[8];
	GetBoundingBox(bbx[0],bbx[1]);
	corner[0].Set(bbx[0].x(),bbx[0].y(),bbx[0].z());
	corner[1].Set(bbx[0].x(),bbx[0].y(),bbx[1].z());
	corner[2].Set(bbx[0].x(),bbx[1].y(),bbx[0].z());
	corner[3].Set(bbx[0].x(),bbx[1].y(),bbx[1].z());
	corner[4].Set(bbx[1].x(),bbx[0].y(),bbx[0].z());
	corner[5].Set(bbx[1].x(),bbx[0].y(),bbx[1].z());
	corner[6].Set(bbx[1].x(),bbx[1].y(),bbx[0].z());
	corner[7].Set(bbx[1].x(),bbx[1].y(),bbx[1].z());

	if(YsDnm::IsPointVisible(8,corner,viewShadowModelTfm,nearZ,farZ,tanFov)==YSTRUE)
	{
		DrawSimpleShadow(pos,att,terOrg,terNom,lightDir);
	}
}

////////////////////////////////////////////////////////////

YsDnmNode::YsDnmNode()
{
	childNameTag=NULL;
	Initialize();
}

void YsDnmNode::Initialize()
{
	if(childNameTag!=NULL)
	{
		delete childNameTag;
		childNameTag=NULL;
	}
	stateList.Set(0,NULL);
	parent=NULL;
	children.Set(0,NULL);

	dnmClassId=0;                    // 2003/03/25
	nodeName.Set("");
	posInParent=YsOrigin();          // 2003/03/25
	attInParent=YsZeroAtt();         // 2003/03/25
	localRotationCenter=YsOrigin();  // 2003/03/25
	localRotationAxis=YsOrigin();    // 2010/02/14
	currentState.show=YSTRUE;        // 2003/03/25
	currentState.relPos=YsOrigin();  // 2003/03/25
	currentState.relAtt=YsZeroAtt(); // 2003/03/25

	YsShellVisual::CleanUp();
}

void YsDnmNode::SetState(int sta)
{
	if(0<=sta && sta<stateList.GetN())
	{
		currentState=stateList[sta];
	}
	else
	{
		currentState.show=YSTRUE;
		currentState.relPos=YsOrigin();
		currentState.relAtt=YsZeroAtt();
	}
}

void YsDnmNode::SetShow(YSBOOL show)
{
	currentState.show=show;
}

void YsDnmNode::InterpolateState(int staA,int staB,const double &param)
{
	if(0<=staA && staA<stateList.GetN() && 0<=staB && staB<stateList.GetN())
	{
		YSCOORDSYSMODEL prevCoordSys;
		prevCoordSys=YsCoordSysModel;
		YsCoordSysModel=YSBLUEIMPULSE;

		YsVec3 evA,uvA,evB,uvB,ev,uv;
		evA=stateList[staA].relAtt.GetForwardVector();
		uvA=stateList[staA].relAtt.GetUpVector();
		evB=stateList[staB].relAtt.GetForwardVector();
		uvB=stateList[staB].relAtt.GetUpVector();

		ev=evA*(1.0-param)+evB*param;
		uv=uvA*(1.0-param)+uvB*param;
		currentState.relAtt.SetTwoVector(ev,uv);

		YsVec3 posA,posB,pos;
		posA=stateList[staA].relPos;
		posB=stateList[staB].relPos;
		pos=posA*(1.0-param)+posB*param;
		currentState.relPos=pos;

		if(param<0.01)
		{
			currentState.show=stateList[staA].show;
		}
		else if(param>0.99)
		{
			currentState.show=stateList[staB].show;
		}
		else
		{
			if(stateList[staA].show==YSTRUE || stateList[staB].show==YSTRUE)
			{
				currentState.show=YSTRUE;
			}
			else
			{
				currentState.show=YSFALSE;
			}
		}

		YsCoordSysModel=prevCoordSys;
	}
	else
	{
		SetState(0);
	}
}

void YsDnmNode::SetHeading(const double &h)
{
	currentState.relAtt.SetH(h);
}

void YsDnmNode::SetPitch(const double &p)
{
	currentState.relAtt.SetP(p);
}

void YsDnmNode::SetBank(const double &b)
{
	currentState.relAtt.SetH(b);
}

////////////////////////////////////////////////////////////

void YsDnmPackedFile::Initialize(void)
{
	fn[0]=0;
	txt.Set(0,NULL);
}

////////////////////////////////////////////////////////////

YsListAllocator <YsDnmNode> YsDnm::nodeAllocator(16);
YsListAllocator <YsDnmPackedFile> YsDnm::packedFileAllocator;

YsDnm::YsDnm() : nodeList(nodeAllocator), packedFileList(packedFileAllocator)
{
}

YsDnm::~YsDnm()
{
	CleanUp();
	// It is dangerous to call CollectGarbate here.
	// Because nodeAllocator is static, nodeAllocator may be destoyed
	// before a YsDnm at the end of the program.
	// nodeAllocator.CollectGarbage();
}

void YsDnm::CleanUp(void)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		node->dat.CleanUp();
		node->dat.stateList.Set(0,NULL);
		node->dat.children.Set(0,NULL);
	}

	nodeList.CleanUp();
}

YsDnmNode *YsDnm::CreateNode(void)
{
	YsListItem <YsDnmNode> *newNode;
	newNode=nodeList.Create();
	if(newNode!=NULL)
	{
		newNode->dat.Initialize();
		newNode->dat.childNameTag=new YsArray <YsString>;
		newNode->dat.thisInTheList=newNode;
		return &newNode->dat;
	}
	return NULL;
}

void YsDnm::DeleteChildNodeNameTag(void)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		delete node->dat.childNameTag;
		node->dat.childNameTag=NULL;
	}
}

YSRESULT YsDnm::LoadDnm(const char fn[])
{
	YSRESULT res=YSERR;
	FILE *fp=fopen(fn,"r");
	if(fp!=NULL)
	{
		res=LoadDnm(fp);
		fclose(fp);
	}
	return res;
}

YSRESULT YsDnm::LoadDnm(FILE *fp)
{
	YsTextFileInputStream inStream(fp);
	return LoadDnm(inStream,"./dnm.dnm");
}

YSRESULT YsDnm::LoadDnm(YsTextInputStream &inStream,const char fn[])
{
	CleanUp();

	YsString idStr;
	int line=1;
	if(NULL!=inStream.Gets(idStr))
	{
		++line;
		// fseek(fp,0,SEEK_SET); I don't think it's needed.  Why did I write this? 
		idStr.Capitalize();
		if(strncmp(idStr,"DYNAMODEL",9)==0)
		{
			if(BeginReadDnm(fn)==YSOK)
			{
				YsString buf;
				while(NULL!=inStream.Gets(buf))
				{
					line++;
					if(ReadDnmOneLine(buf)!=YSOK)
					{
						printf("Line %d\n",line);
						EndReadDnm();
						goto ERREND;
					}
				}
				if(EndReadDnm()==YSOK)
				{
					return YSOK;
				}
			}
		}
		else if(strncmp(idStr,"SURF",4)==0)
		{
			if(BeginReadDnm(fn)==YSOK)
			{
				currentNode=CreateNode();
				currentNode->nodeName.Set("SHELL");

				currentNode->BeginReadSrf();
				YsString buf;
				while(NULL!=inStream.Gets(buf))
				{
					line++;
					currentNode->ReadSrfOneLine(buf);
				}
				currentNode->EndReadSrf();

				if(EndReadDnm()==YSOK)
				{
					return YSOK;
				}
			}
		}
	}

ERREND:
	return YSERR;
}

YSRESULT YsDnm::BeginReadDnm(const char fn[])
{
	readState=0;
	currentNode=NULL;
	nPackedFileRemain=0;
	dnmVersion=0;

	if(NULL!=fn)
	{
		YsString ful(fn),fil;
		ful.SeparatePathFile(readPath,fil);
	}
	else
	{
		readPath.Set(".");
	}
	return YSOK;
}

const char *const YsDnm::keyWordSource[]=
{
	"DYNAMODEL", // 0
	"DNMVER",    // 1
	"PCK",       // 2
	"SRF",       // 3
	"PLG",       // 4
	"LIN",       // 5
	"CLA",       // 6
	"NST",       // 7
	"STA",       // 8
	"NCH",       // 9
	"REL",       // 10
	"CLD",       // 11
	"FIL",       // 12
	"VTX",       // 13
	"NVT",       // 14
	"COL",       // 15
	"POS",       // 16
	"CNT",       // 17
	"END",       // 18
	"PAX",       // 19
	NULL
};

YsKeyWordList YsDnm::keyWordList;

YSRESULT YsDnm::ReadDnmOneLine(const char str[])
{
	int ac;
	char *av[256],buf[4096];
	strcpy(buf,str);

	if(keyWordList.GetN()==0)
	{
		keyWordList.MakeList(keyWordSource);
	}

	YsCapitalize(buf);

	if(readState==99)
	{
		/* YsDnmString newStr;
		strcpy(newStr.str,str);
		YsStringTail(newStr.str);
		currentPack->txt.Append(newStr); */

		currentPack->txt.Increment();
		currentPack->txt.GetEnd().Set(str);
		currentPack->txt.GetEnd().DeleteTailSpace();

		nPackedFileRemain--;
		if(nPackedFileRemain==0)
		{
			readState=0;
		}
		return YSOK;
	}
	else if(YsArguments(&ac,av,256,buf)==YSOK && ac>0)
	{
		if(readState==0)
		{
			if(av[0][0]=='#')    // 2003/03/25
			{                         // 2003/03/25
				// Comment line       // 2003/03/25
				return YSOK;          // 2003/03/25
			}                         // 2003/03/25

			int cmdId;
			cmdId=keyWordList.GetId(av[0]);
			switch(cmdId)
			{
			case 0:   // "DYNAMODEL", // 0
				// Ignore
				return YSOK;
			case 1:   // "DNMVER",    // 1
				if(ac>=2)
				{
					dnmVersion=atoi(av[1]);
					return YSOK;
				}
				break;
			case 2:   // "PCK",       // 2
				if(ac>=3)
				{
					currentPack=CreatePackedFile(av[1]);
					nPackedFileRemain=atoi(av[2]);
					if(nPackedFileRemain>0)
					{
						readState=99;
					}
					return YSOK;
				}
				break;
			case 3:   // "SRF",       // 3
				if(ac>=2)
				{
					currentNode=CreateNode();
					currentNode->nodeName.Set(av[1]);
					currentNodeType=0;
					return YSOK;
				}
				break;
			case 4:   // "PLG",       // 4
				if(ac>=2)
				{
					currentNode=CreateNode();
					currentNode->nodeName.Set(av[1]);
					currentNodeType=1;
					return YSOK;
				}
				break;
			case 5:   // "LIN",       // 5
				if(ac>=2)
				{
					currentNode=CreateNode();
					currentNode->nodeName.Set(av[1]);
					currentNodeType=2;
					return YSOK;
				}
				break;
			case 6:   // "CLA",       // 6
				if(ac>=2 && currentNode!=NULL)
				{
					currentNode->dnmClassId=atoi(av[1]);
					return YSOK;
				}
				break;
			case 7:   // "NST",       // 7
				// Ignore
				return YSOK;
			case 8:   // "STA",       // 8
				if(ac>=8 && currentNode!=NULL)
				{
					YsDnmState sta;
					sta.relPos.Set(atof(av[1]),atof(av[2]),atof(av[3]));
					sta.relAtt.Set(atof(av[4])*YsPi/32768,atof(av[5])*YsPi/32768,atof(av[6])*YsPi/32768);
					sta.show=(atoi(av[7])==0 ? YSFALSE : YSTRUE);
					currentNode->stateList.Append(sta);
					return YSOK;
				}
				break;
			case 9:   // "NCH",       // 9
				// Nothing to do
				return YSOK;
			case 10:  // "REL",       // 10
				// Only "DEP" is accepted, thus, ignored
				return YSOK;
			case 11:  // "CLD",       // 11
				if(currentNode!=NULL && ac>=2)
				{
					/* YsDnmString str;
					strcpy(str.str,av[1]);
					currentNode->childNameTag->Append(str); */

					currentNode->childNameTag->Increment();
					currentNode->childNameTag->GetEnd().Set(av[1]);
					return YSOK;
				}
				break;
			case 12:  // "FIL",       // 12
				if(currentNode!=NULL && ac>=2)
				{
					if(currentNodeType==0)
					{
						YSBOOL isPacked;
						YsListItem <YsDnmPackedFile> *packed;
						isPacked=YSFALSE;
						packed=NULL;
						while((packed=packedFileList.FindNext(packed))!=NULL)
						{
							if(strcmp(packed->dat.fn,av[1])==0)
							{
								int i;
								// YsPrintf("Packed File in DNM : %s\n",av[1]);
								currentNode->BeginReadSrf();
								for(i=0; i<packed->dat.txt.GetN(); i++)
								{
									currentNode->ReadSrfOneLine(packed->dat.txt[i]);
								}
								currentNode->EndReadSrf();
								isPacked=YSTRUE;
								break;
							}
						}
						if(isPacked!=YSTRUE)
						{
							YsString ful;
							ful.MakeFullPathName(readPath,av[1]);
							if(currentNode->LoadSrf(ful)!=YSOK)
							{
								YsPrintf("Cannot read %s\n",ful.Txt());
								readState=-1;
								return YSERR;
							}
						}
					}
					return YSOK;
				}
				break;
			case 13:  // "VTX",       // 13
				if(currentNode!=NULL && ac==7)
				{
					if(currentNodeType==2)
					{
						double l;
						YsVec3 p1,p2,u,v,w,vv,ww;
						p1.Set(atof(av[1]),atof(av[2]),atof(av[3]));
						p2.Set(atof(av[4]),atof(av[5]),atof(av[6]));
						u=p2-p1;
						u.Normalize();
						v=u.GetArbitraryPerpendicularVector();
						v.Normalize();
						w=u^v;
						l=(p2-p1).GetLength();
						vv=v*0.015;  // vv=v*l/50.0;  2003/03/26
						ww=w*0.015;  // ww=w*l/50.0;  2003/03/26

						YsShellVertexHandle vtHd[8],quad[4];
						YsShellPolygonHandle plHd;
						vtHd[0]=currentNode->AddVertexH(p1+vv+ww);
						vtHd[1]=currentNode->AddVertexH(p1+vv-ww);
						vtHd[2]=currentNode->AddVertexH(p1-vv-ww);
						vtHd[3]=currentNode->AddVertexH(p1-vv+ww);
						vtHd[4]=currentNode->AddVertexH(p2+vv+ww);
						vtHd[5]=currentNode->AddVertexH(p2+vv-ww);
						vtHd[6]=currentNode->AddVertexH(p2-vv-ww);
						vtHd[7]=currentNode->AddVertexH(p2-vv+ww);

						quad[0]=vtHd[0];
						quad[1]=vtHd[1];
						quad[2]=vtHd[2];
						quad[3]=vtHd[3];
						plHd=currentNode->AddPolygonH(4,quad);
						currentNode->SetColorOfPolygon(plHd,YsWhite());
						currentNode->SetNormalOfPolygon(plHd,-u);

						quad[0]=vtHd[7];
						quad[1]=vtHd[6];
						quad[2]=vtHd[5];
						quad[3]=vtHd[4];
						plHd=currentNode->AddPolygonH(4,quad);
						currentNode->SetColorOfPolygon(plHd,YsWhite());
						currentNode->SetNormalOfPolygon(plHd,u);

						quad[0]=vtHd[1];
						quad[1]=vtHd[0];
						quad[2]=vtHd[4];
						quad[3]=vtHd[5];
						plHd=currentNode->AddPolygonH(4,quad);
						currentNode->SetColorOfPolygon(plHd,YsWhite());
						currentNode->SetNormalOfPolygon(plHd,v);

						quad[0]=vtHd[2];
						quad[1]=vtHd[1];
						quad[2]=vtHd[5];
						quad[3]=vtHd[6];
						plHd=currentNode->AddPolygonH(4,quad);
						currentNode->SetColorOfPolygon(plHd,YsWhite());
						currentNode->SetNormalOfPolygon(plHd,-w);

						quad[0]=vtHd[3];
						quad[1]=vtHd[2];
						quad[2]=vtHd[6];
						quad[3]=vtHd[7];
						plHd=currentNode->AddPolygonH(4,quad);
						currentNode->SetColorOfPolygon(plHd,YsWhite());
						currentNode->SetNormalOfPolygon(plHd,-v);

						quad[0]=vtHd[0];
						quad[1]=vtHd[3];
						quad[2]=vtHd[7];
						quad[3]=vtHd[4];
						plHd=currentNode->AddPolygonH(4,quad);
						currentNode->SetColorOfPolygon(plHd,YsWhite());
						currentNode->SetNormalOfPolygon(plHd,w);

						currentNode->RecomputeNormalAndFlipDirection();
						currentNode->SortPolygonByColor();
						currentNode->PrecomputePolygCenterAfterSortingByColor();
					}
					return YSOK;
				}
				break;
			case 14:  //  "NVT",       // 14
				if(currentNode!=NULL && ac>=2)
				{
					nVtxRemain=atoi(av[1]);
					if(nVtxRemain>0)
					{
						readState=1;
					}
					return YSOK;
				}
				break;
			case 15:  // "COL",       // 15
				if(currentNode!=NULL && ac>=2)
				{
					YsColor col;
					YsShellPolygonHandle plHd;
					if(ac>=4)
					{
						col.SetIntRGB(atoi(av[1]),atoi(av[2]),atoi(av[3]));
					}
					else
					{
						col.Set15BitRGB(atoi(av[1]));
					}
					plHd=NULL;
					while((plHd=currentNode->FindNextPolygon(plHd))!=NULL)
					{
						currentNode->SetColorOfPolygon(plHd,col);
					}
					return YSOK;
				}
				break;
			case 16:  // "POS",       // 16
				if(currentNode!=NULL && ac>=7)
				{
					double x,y,z,h,p,b;
					x=atof(av[1]);
					y=atof(av[2]);
					z=atof(av[3]);
					h=atof(av[4])*YsPi/32768;
					p=atof(av[5])*YsPi/32768;
					b=atof(av[6])*YsPi/32768;
					currentNode->posInParent.Set(x,y,z);
					currentNode->attInParent.Set(h,p,b);
					return YSOK;
				}
				break;
			case 17:  // "CNT",       // 17
				if(currentNode!=NULL && ac>=4)
				{
					double x,y,z;
					x=atof(av[1]);
					y=atof(av[2]);
					z=atof(av[3]);
					currentNode->localRotationCenter.Set(x,y,z);
					return YSOK;
				}
				break;
			case 18:  // "END",       // 18
				if(currentNode!=NULL)
				{
					currentNode=NULL;
				}
				else
				{
					readState=999;
				}
				return YSOK;
			case 19:  // "PAX",       // 19
				if(currentNode!=NULL && ac>=4)
				{
					double x,y,z;
					x=atof(av[1]);
					y=atof(av[2]);
					z=atof(av[3]);
					currentNode->localRotationAxis.Set(x,y,z);
					return YSOK;
				}
				break;
			}
		}
		else if(readState==1)
		{
			YsShellVertexHandle vtHd;
			if(ac>=4)
			{
				YsVec3 pos;
				pos.Set(atof(av[1]),atof(av[2]),atof(av[3]));
				currentNode->AddVertexH(pos);
				nVtxRemain--;
				if(nVtxRemain==0)
				{
					vtHd=NULL;
					YsArray <YsShellVertexHandle,256> plVtHd;
					YsArray <YsVec3> plVtPos;
					while((vtHd=currentNode->FindNextVertex(vtHd))!=NULL)
					{
						plVtHd.Append(vtHd);
						currentNode->GetVertexPosition(pos,vtHd);
						plVtPos.Append(pos);
					}

					YsShellPolygonHandle plHd;
					plHd=currentNode->AddPolygonH(plVtHd.GetN(),plVtHd);
					currentNode->GetPolygon(plHd)->freeAttribute0=YsShellVisual::PLG_TWOSIDED;

					YsVec3 nom;
					YsGetAverageNormalVector(nom,plVtPos.GetN(),plVtPos);
					nom.Normalize();
					currentNode->SetNormalOfPolygon(plHd,nom);

					if(YsCheckConvexByAngle3(plVtPos.GetN(),plVtPos)!=YSTRUE)
					{
						currentNode->GetPolygon(plHd)->freeAttribute0|=YsShellVisual::PLG_CONCAVE;
					}

					currentNode->RecomputeNormalAndFlipDirection();
					currentNode->SortPolygonByColor();
					currentNode->PrecomputePolygCenterAfterSortingByColor();

					readState=0;
				}
				return YSOK;
			}
			else
			{
				YsPrintf("Error during reading polygon node.\n");
				YsPrintf("  >> %s\n",str);
				readState=-1;
				return YSERR;
			}
		}
		else if(readState==999)
		{
			return YSOK;
		}

		YsPrintf("Error during reading DNM\n");
		YsPrintf("  >> %s\n",str);
		return YSERR;
	}
	return YSOK;
}

YSRESULT YsDnm::EndReadDnm(void)
{
	MakeConnectionAndRootNodeListByChildNameTag();
	DeleteChildNodeNameTag();
	packedFileList.CleanUp();
	packedFileAllocator.CollectGarbage();
	ClearState();

	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		node->dat.Encache();
	}

	nodeList.Encache();

	ComputeBoundingBox();

	return YSOK;
}

YSRESULT YsDnm::SaveDnm(const char fn[]) const
{
	FILE *fp;
	fp=fopen(fn,"w");
	if(fp!=NULL)
	{
		YsTextFileOutputStream textOut(fp);
		YSRESULT res=SaveDnm(textOut);
		fclose(fp);
		return res;
	}
	return YSERR;
}

YSRESULT YsDnm::SaveDnm(YsTextOutputStream &textOut) const
{
	textOut.Printf("DYNAMODEL\n");
	textOut.Printf("DNMVER %d\n",dnmVersion);

	if(YSOK!=SaveDnmAddPackAll(textOut))
	{
		goto ERREND;
	}

	YsListItem <YsDnmNode> *dnmNode;
	dnmNode=NULL;
	while(NULL!=(dnmNode=nodeList.FindNext(dnmNode)))
	{
		int i;

		textOut.Printf("SRF \"%s\"\n",dnmNode->dat.nodeName.Txt());
		textOut.Printf("FIL \"%s.srf\"\n",dnmNode->dat.nodeName.Txt());
		textOut.Printf("CLA %d\n",dnmNode->dat.dnmClassId);

		textOut.Printf("NST %d\n",dnmNode->dat.stateList.GetN());
		forYsArray(i,dnmNode->dat.stateList)
		{
			textOut.Printf("STA %lf %lf %lf %d %d %d %d\n",
				dnmNode->dat.stateList[i].relPos.x(),
				dnmNode->dat.stateList[i].relPos.y(),
				dnmNode->dat.stateList[i].relPos.z(),
				(int)(dnmNode->dat.stateList[i].relAtt.h()*32768.0/YsPi),
				(int)(dnmNode->dat.stateList[i].relAtt.p()*32768.0/YsPi),
				(int)(dnmNode->dat.stateList[i].relAtt.b()*32768.0/YsPi),
				(YSTRUE==dnmNode->dat.stateList[i].show ? 1 : 0));
		}

		textOut.Printf("POS %lf %lf %lf %d %d %d\n",
		    dnmNode->dat.posInParent.x(),
		    dnmNode->dat.posInParent.y(),
		    dnmNode->dat.posInParent.z(),
		    (int)(dnmNode->dat.attInParent.h()*32768.0/YsPi),
		    (int)(dnmNode->dat.attInParent.p()*32768.0/YsPi),
		    (int)(dnmNode->dat.attInParent.b()*32768.0/YsPi));

		textOut.Printf("CNT %lf %lf %lf\n",
		    dnmNode->dat.localRotationCenter.x(),
		    dnmNode->dat.localRotationCenter.y(),
		    dnmNode->dat.localRotationCenter.z());
		textOut.Printf("PAX %lf %lf %lf\n",
		    dnmNode->dat.localRotationAxis.x(),
		    dnmNode->dat.localRotationAxis.y(),
		    dnmNode->dat.localRotationAxis.z());
		textOut.Printf("REL DEP\n");

		textOut.Printf("NCH %d\n",dnmNode->dat.children.GetN());
		forYsArray(i,dnmNode->dat.children)
		{
			textOut.Printf("CLD \"%s\"\n",dnmNode->dat.children[i]->nodeName.Txt());
		}
		textOut.Printf("END\n");
	}

	return YSOK;
ERREND:
	return YSERR;
}

YSRESULT YsDnm::SaveDnmAddPackAll(YsTextOutputStream &textOut) const
{
	YsListItem <YsDnmNode> *dnmNode;
	dnmNode=NULL;
	while(NULL!=(dnmNode=nodeList.FindNext(dnmNode)))
	{
		if(YSOK!=SaveDnmAddPack(textOut,&dnmNode->dat))
		{
			return YSERR;
		}
	}
	return YSOK;
}

YSRESULT YsDnm::SaveDnmAddPack(YsTextOutputStream &textOut,YsDnmNode *node) const
{
	YsTextMemoryOutputStream tmpTextOut;

	if(node->SaveSrf(tmpTextOut)==YSOK)
	{
		const int numLine=tmpTextOut.GetNumLine();
		textOut.Printf("PCK \"%s.srf\" %d\n",node->nodeName.Txt(),numLine);

		YsTextMemoryOutputStreamStringHandle stHd=NULL;
		while(NULL!=(stHd=tmpTextOut.FindNextLine(stHd)))
		{
			textOut.Printf("%s\n",tmpTextOut.GetLine(stHd)->Txt());
		}
		textOut.Printf("\n\n\n");

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsDnm::MakeConnectionAndRootNodeListByChildNameTag(void)
{
	YsListItem <YsDnmNode> *node;

	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		node->dat.parent=NULL;
		node->dat.children.Set(0,NULL);
	}

	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		YsListItem <YsDnmNode> *seekChild;
		seekChild=NULL;
		while((seekChild=nodeList.FindNext(seekChild))!=NULL)
		{
			int i;
			for(i=0; i<node->dat.childNameTag->GetN(); i++)
			{
				if(strcmp((*node->dat.childNameTag)[i],seekChild->dat.nodeName)==0)
				{
					YSBOOL isLoop;         // Loop check 2003/03/26
					YsDnmNode *loopCheck;
					isLoop=YSFALSE;
					loopCheck=&node->dat;
					while(loopCheck!=NULL)
					{
						if(loopCheck==&seekChild->dat)
						{
							YsPrintf("Loop in DNM file!\n");
							isLoop=YSTRUE;
							break;
						}
						loopCheck=loopCheck->parent;
					}

					if(isLoop!=YSTRUE)
					{
						seekChild->dat.parent=&node->dat;
						node->dat.children.Append(&seekChild->dat);
					}
					break;
				}
			}
		}
	}

	rootNode.Set(0,NULL);
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		if(node->dat.parent==NULL)
		{
			rootNode.Append(&node->dat);
		}
	}

	if(dnmVersion==0)  // Mimic the bug of the old DNM class.  2003/03/26
	{
		for(YSSIZE_T i=rootNode.GetN()-1; i>0; i--)
		{
			rootNode[0]->children.Append(rootNode[i]);
			rootNode[i]->parent=rootNode[0];
			rootNode.Delete(i);
		}
	}

	return YSOK;
}

YsDnmPackedFile *YsDnm::CreatePackedFile(const char fn[])
{
	YsListItem <YsDnmPackedFile> *newPack;
	newPack=packedFileList.Create();
	if(newPack!=NULL)
	{
		newPack->dat.Initialize();
		strcpy(newPack->dat.fn,fn);
		return &newPack->dat;
	}
	return NULL;
}

void YsDnm::ComputeBoundingBox(void)
{
	int state,i;
	YsBoundingBoxMaker3 makeBbx;

	makeBbx.Begin();
	for(state=0; state<3; state++)
	{
		SetState(state);
		for(i=0; i<rootNode.GetN(); i++)
		{
			ComputeBoundingBox(makeBbx,rootNode[i],YsIdentity4x4());
		}
	}
	makeBbx.Get(bbx[0],bbx[1]);
}

void YsDnm::ComputeBoundingBox(YsBoundingBoxMaker3 &bbx,YsDnmNode *node,const YsMatrix4x4 &current)
{
	YsMatrix4x4 newCurrent;
	newCurrent=current;

	newCurrent.Translate(node->posInParent);

	newCurrent.RotateXZ(node->attInParent.h());
	newCurrent.RotateZY(node->attInParent.p());
	newCurrent.RotateXY(node->attInParent.b());

	newCurrent.Translate(node->currentState.relPos);

	newCurrent.RotateXZ(node->currentState.relAtt.h());
	newCurrent.RotateZY(node->currentState.relAtt.p());
	newCurrent.RotateXY(node->currentState.relAtt.b());

	newCurrent.Translate(-node->localRotationCenter);

	YsVec3 nodeBbx[2],corner[8];;
	node->GetBoundingBox(nodeBbx[0],nodeBbx[1]);

	corner[0].Set(nodeBbx[0].x(),nodeBbx[0].y(),nodeBbx[0].z());
	corner[1].Set(nodeBbx[0].x(),nodeBbx[0].y(),nodeBbx[1].z());
	corner[2].Set(nodeBbx[0].x(),nodeBbx[1].y(),nodeBbx[0].z());
	corner[3].Set(nodeBbx[0].x(),nodeBbx[1].y(),nodeBbx[1].z());
	corner[4].Set(nodeBbx[1].x(),nodeBbx[0].y(),nodeBbx[0].z());
	corner[5].Set(nodeBbx[1].x(),nodeBbx[0].y(),nodeBbx[1].z());
	corner[6].Set(nodeBbx[1].x(),nodeBbx[1].y(),nodeBbx[0].z());
	corner[7].Set(nodeBbx[1].x(),nodeBbx[1].y(),nodeBbx[1].z());

	int i;
	for(i=0; i<8; i++)
	{
		newCurrent.Mul(corner[i],corner[i],1.0);
		bbx.Add(corner[i]);
	}

	for(i=0; i<node->children.GetN(); i++)
	{
		ComputeBoundingBox(bbx,node->children[i],newCurrent);
	}
}

void YsDnm::DrawVisual
    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,
     const double &nearZ,const double &farZ,const double &tanFov,unsigned drawFlag) const
{
	YsMatrix4x4 viewModelTfm;
	viewModelTfm=viewTfm;
	viewModelTfm.Translate(pos);
	viewModelTfm.RotateXZ(att.h());
	viewModelTfm.RotateZY(att.p());
	viewModelTfm.RotateXY(att.b());

	if(IsBbxVisible(bbx,viewModelTfm,nearZ,farZ,tanFov)==YSTRUE)
	{
		DrawWithFullKnowledge(viewTfm,pos,att,drawFlag);
	}
}

void YsDnm::DrawShadowVisual
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir,
	     const double &nearZ,const double &farZ,const double &tanFov) const
{
	YsMatrix4x4 prjMat(YSFALSE);
	YsMakePlaneProjectionMatrix(prjMat,terOrg,terNom,lightDir);

	YsMatrix4x4 viewShadowModelTfm(YSFALSE);
	viewShadowModelTfm=viewTfm;
	viewShadowModelTfm*=prjMat;
	// viewShadowModelTfm.Translate(0.0,elv,0.0);
	// viewShadowModelTfm.Scale(1.0,0.0,1.0);
	viewShadowModelTfm.Translate(pos);
	viewShadowModelTfm.RotateXZ(att.h());
	viewShadowModelTfm.RotateZY(att.p());
	viewShadowModelTfm.RotateXY(att.b());

	YsVec3 corner[8];
	corner[0].Set(bbx[0].x(),bbx[0].y(),bbx[0].z());
	corner[1].Set(bbx[0].x(),bbx[0].y(),bbx[1].z());
	corner[2].Set(bbx[0].x(),bbx[1].y(),bbx[0].z());
	corner[3].Set(bbx[0].x(),bbx[1].y(),bbx[1].z());
	corner[4].Set(bbx[1].x(),bbx[0].y(),bbx[0].z());
	corner[5].Set(bbx[1].x(),bbx[0].y(),bbx[1].z());
	corner[6].Set(bbx[1].x(),bbx[1].y(),bbx[0].z());
	corner[7].Set(bbx[1].x(),bbx[1].y(),bbx[1].z());

	if(IsPointVisible(8,corner,viewShadowModelTfm,nearZ,farZ,tanFov)==YSTRUE)
	{
		DrawSimpleShadow(pos,att,terOrg,terNom,lightDir);
	}
}

void YsDnm::ClearState(void)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		node->dat.currentState.relPos=YsOrigin();
		node->dat.currentState.relAtt=YsZeroAtt();
		node->dat.currentState.show=YSTRUE;
	}
}

void YsDnm::SetState(int sta)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		node->dat.SetState(sta);
	}
}

void YsDnm::SetClassState(int clId,int sta)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		if(node->dat.dnmClassId==clId)
		{
			node->dat.SetState(sta);
		}
	}
}

void YsDnm::SetPapiColor(const double &angle)
{
	if(rootNode.GetN()>0 && rootNode[0]!=NULL)
	{
		// On course
		// White White Red Red  (RUNWAY)  Red Red White White

		//                1.5 deg          3 deg             4.5 deg
		// t              0                0.5               1.0
		// Light [0]      Red     Red      Red      Red      White       (0.75 to 1.0)
		// Light [1]      Red     Red      Red      White    White       (0.5  to 0.75)
		// Light [2]      Red     Red      White    White    White       (0.25 to 0.5)
		// Light [3]      Red     White    White    White    White       (0.0  to 0.25)

		YsShellPolygonHandle plHd;
		YsColor col[4];
		int i,plId;
		double t,t0,gb;
		t=(YsRadToDeg(angle)-1.5)/3.0;

		for(i=0; i<4; i++)
		{
			t0=0.75-0.25*(double)i;
			gb=YsBound((t-t0)/0.25,0.0,1.0);
			col[i].SetDoubleRGB(1.0,gb,gb);
		}

		plHd=NULL;
		plId=0;
		while(NULL!=(plHd=rootNode[0]->FindNextPolygon(plHd)))
		{
			rootNode[0]->SetColorOfPolygon(plHd,col[4*plId/rootNode[0]->GetNumPolygon()]);
			plId++;
		}

		rootNode[0]->flags|=YsShellVisual::SHL_DONOTCACHE;
		rootNode[0]->flags|=YsShellVisual::SHL_DRAWPOINT;
	}
}

void YsDnm::SetVasiColor(const double &angle)
{
	if(rootNode.GetN()>0 && rootNode[0]!=NULL)
	{
		//                2.5 deg          3.5deg
		// t              0                1.0
		//                Red              White

		YsShellPolygonHandle plHd;
		YsColor col;
		double t,gb;
		t=(YsRadToDeg(angle)-2.5);

		gb=YsBound(t,0.0,1.0);
		col.SetDoubleRGB(1.0,gb,gb);

		plHd=NULL;
		while(NULL!=(plHd=rootNode[0]->FindNextPolygon(plHd)))
		{
			rootNode[0]->SetColorOfPolygon(plHd,col);
		}

		rootNode[0]->flags|=YsShellVisual::SHL_DONOTCACHE;
		rootNode[0]->flags|=YsShellVisual::SHL_DRAWPOINT;
	}
}

void YsDnm::InterpolateClassState(int clId,int staA,int staB,const double &param)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		if(node->dat.dnmClassId==clId)
		{
			node->dat.InterpolateState(staA,staB,param);
		}
	}
}

void YsDnm::SetShow(int clId,YSBOOL show)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		if(node->dat.dnmClassId==clId)
		{
			node->dat.SetShow(show);
		}
	}
}

void YsDnm::SetClassHeading(int clId,const double &h)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		if(node->dat.dnmClassId==clId)
		{
			node->dat.SetHeading(h);
		}
	}
}

void YsDnm::SetClassPitch(int clId,const double &p)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		if(node->dat.dnmClassId==clId)
		{
			node->dat.SetPitch(p);
		}
	}
}

void YsDnm::SetClassBank(int clId,const double &b)
{
	YsListItem <YsDnmNode> *node;
	node=NULL;
	while((node=nodeList.FindNext(node))!=NULL)
	{
		if(node->dat.dnmClassId==clId)
		{
			node->dat.SetBank(b);
		}
	}
}

void YsDnm::GetBoundingBox(YsVec3 b[2]) const
{
	b[0]=bbx[0];
	b[1]=bbx[1];
}

int YsDnm::FindDnmNodeIdByTag(const char tag[]) const
{
	int id;
	YsListItem <YsDnmNode> *node;
	YsString str;
	node=NULL;
	id=0;
	while(NULL!=(node=nodeList.FindNext(node)))
	{
		str=node->dat.nodeName;
		str.Capitalize();
		if(strcmp(str,tag)==0)
		{
			return id;
		}
		id++;
	}
	return -1;
}

void YsDnm::SetNodeHeadingPitch(int nodeId,const double &h,const double &p)
{
	YsListItem <YsDnmNode> *node;
	node=nodeList.GetItemFromId(nodeId);
	if(node!=NULL)
	{
		node->dat.SetHeading(h);
		node->dat.SetPitch(p);
	}
}

/*
YSRESULT GpShellGroup::MoveDnm
    (GpShellHandle slHd,int state1,int state2,const double &param,const YsMatrix4x4 &current)
{
	YsMatrix4x4 local,Tp,Rp,Rl,Tl,Rm,Tm;
	YsMatrix4x4 iTl;
	GpShell *shl;

	// Tp : Center in parent coordsys
	// Rp : Rotation in parent
	// Tl : Rotation center in local coordsys
	// Rl : Rotation in local
	// Rm : Rotation by motion
	// Tm : Transaltion by motion

	if(slHd!=NULL && (shl=GetEditableShell(slHd))!=NULL)
	{
		Tp.Translate(shl->dnmInfo.centerInParent);

		double h,p,b;
		VectorToAngle(h,p,b,shl->dnmInfo.evInParent,shl->dnmInfo.uvInParent);
		Rp.Initialize();
		Rp.RotateXZ(h);
		Rp.RotateZY(p);
		Rp.RotateXY(b);

		if(0<=state1 && state1<shl->dnmInfo.nState &&
		   0<=state2 && state2<shl->dnmInfo.nState)
		{
			double h,p,b;
			YsVec3 trans,ev,uv;
			trans=shl->dnmInfo.dnmState[state1].relPos*(1.0-param)
			     +shl->dnmInfo.dnmState[state2].relPos*param;
			Tm.Translate(trans);

			ev=shl->dnmInfo.dnmState[state1].relEv*(1.0-param)
			  +shl->dnmInfo.dnmState[state2].relEv*param;
			uv=shl->dnmInfo.dnmState[state1].relUv*(1.0-param)
			  +shl->dnmInfo.dnmState[state2].relUv*param;
			VectorToAngle(h,p,b,ev,uv);
			Rm.Initialize();
			Rm.RotateXZ(h);
			Rm.RotateZY(p);
			Rm.RotateXY(b);

			if(state1==state2 || param<0.01)
			{
				shl->dnmInfo.disp=shl->dnmInfo.dnmState[state1].disp;
			}
			else if(param>0.09)
			{
				shl->dnmInfo.disp=shl->dnmInfo.dnmState[state2].disp;
			}
			else if(shl->dnmInfo.dnmState[state1].disp==YSTRUE ||
			        shl->dnmInfo.dnmState[state2].disp==YSTRUE)
			{
				shl->dnmInfo.disp=YSTRUE;
			}
			else
			{
				shl->dnmInfo.disp=YSFALSE;
			}
		}
		else
		{
			Tm.Initialize();
			Rm.Initialize();
			shl->dnmInfo.disp=YSTRUE;
		}

		Tl.Translate(shl->dnmInfo.centerInLocal);
		Rl.Initialize();

		iTl=Tl;
		iTl.Invert();

		local=current*Tp*Rp*Tm*Rm*Rl*iTl;
		shl->shl.SetMatrix(local);

		MoveDnm(shl->children,state1,state2,param,local);
		MoveDnm(shl->thisInTheList->Next(),state1,state2,param,current);
	}
	return YSOK;
}
*/

