#ifndef YSDNM_IS_INCLUDED
#define YSDNM_IS_INCLUDED
/* { */

/* Use of free attributes

See wingepoly/gpshell.h

*/


enum
{
	YSDNM_DRAWOPAQUE=1,
	YSDNM_DRAWTRANSPARENT=2,

	YSDNM_DRAWALL=3
};

class YsShellVisualLight
{
public:
	YsShellPolygonHandle plHd;
	YsVec3 cen,nom;
	double rad;
	YsColor col;
};

class YsShellVisual : public YsShell
{
public:
	enum VTXFLAG
	{
		VTX_ROUND=1,
		VTX_NODECONST=2   // Not used in YsShellVisual
	};
	enum PLGFLAG
	{
		PLG_NOSHADING=1,
		PLG_TWOSIDED=2,
		PLG_CONCAVE=4,
		PLG_INVISINONOPENGL=8,  // <- Don't draw when transparency is off.
		// bit 4-7 unused
		// bit 8-15 Transparency
		PLG_LIGHT=0x100000
		// bit 17-31 unused
	};
	enum SHELLFLAG
	{
		SHL_DONOTCACHE=1,
		SHL_DRAWPOINT=2
	};
	YsShellVisual();
	~YsShellVisual();
	const YsShellVisual &operator=(const YsShellVisual &from);

	static int globalControl_lightPointSize;
	static unsigned globalControl_transparency;  // 0:No transparency  255:Use Trancparency
	static YSBOOL globalControl_useBiSubTree;
	static YSBOOL globalControl_useDisplayList;
	static YSBOOL globalControl_triangulateAll;
	static YSBOOL globalControl_antiAliasing;

	unsigned int flags;
	unsigned int openGlListId;
	class YsDnmExternalVertexBufferLink *extBufLink;

	// ReadSrf Extension
protected:
	YsArray <YsShellPolygonHandle> *noShadPlg;
	YsArray <YsShellVertexHandle> *roundVtx;
	YsArray <unsigned> *plgAhpla;
	YsArray <unsigned> *invisiNonOpenGL;
	YsArray <unsigned> *plgLightFlag;
	void ReadSrfPostProcess(void);

	void AllocateD3dExternalVertexBufferLink(void);
	void FreeD3dExternalVertexBufferLink(void);
	void ReleaseD3dExternalVertexBuffer(void);

public:
	YsArray <YsVec3> plgCenter;
	YsArray <YsShellVisualLight> light;

	void CleanUp(void);

	YSRESULT LoadSrf(const char fn[]);
	YSRESULT LoadSrf(FILE *fp);
	YSRESULT LoadSrf(YsTextInputStream &inStream);
	YSRESULT BeginReadSrf(void);
	YSRESULT ReadSrfOneLine(const char str[]);
	YSRESULT EndReadSrf(void);
	void CheckLeak(void);

	YSRESULT SaveSrf(const char fn[]) const;
	YSRESULT SaveSrf(YsTextOutputStream &textOut) const;

	void RecomputeNormalAndFlipDirection(void);
	void TriangulateAll(void);
	void SimpleTriangulation(YsShellPolygonHandle plHd);
	void SortPolygonByColor(void);
	void PrecomputePolygCenterAfterSortingByColor(void);

	// Draw shell without transforming
	void Draw(void) const;
	// Draw shell without transforming, knowing current viewTfm*modelTfm
	void Draw(const YsMatrix4x4 &viewModelTfm) const;

	// Draw shell after rotating by att and translating by pos
	void Draw(const YsVec3 &pos,const YsAtt3 &att) const;

	// Draw shell after rotating by att and translating by pos, knowing viewing frustum.
	void DrawVisual
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,
	     const double &nearZ,const double &farZ,const double &tanFov,unsigned int drawFlag) const;
	void DrawVisual(const YsMatrix4x4 &viewTfm,unsigned int drawFlag) const;
	void DrawLightVisual(const YsMatrix4x4 &viewTfm) const;

	// Draw shell for making display list
	void DrawForMakingOpenGlDisplayList(unsigned int listId);

	// Draw shell after rotating by att and translating by pos, knowing viewing frustum.
	// In Non-OpenGL version, the shell will be drawn as low priority polygons.
	void DrawVisualLowPriorityMode
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,
	     const double &nearZ,const double &farZ,const double &tanFov) const;

	// Draw shell shadow after rotating by att and translating by pos, knowing viewing frustum.
	void DrawShadowVisual
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir,
	     const double &nearZ,const double &farZ,const double &tanFov) const;

	void DrawSimpleShadow(const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir) const;
};


////////////////////////////////////////////////////////////

class YsDnmState
{
public:
	YSBOOL show;
	YsVec3 relPos;
	YsAtt3 relAtt;
};

class YsDnmNode : public YsShellVisual
{
friend class YsDnm;
protected:
	YsArray <YsString> *childNameTag;
public:
	YsDnmNode();
	void Initialize();

	YsArray <YsDnmState> stateList;
	YsDnmState currentState;

	YsString nodeName;
	YsDnmNode *parent;
	YsArray <YsDnmNode *> children;

	int dnmClassId;
	YsVec3 posInParent;
	YsAtt3 attInParent;
	YsVec3 localRotationCenter,localRotationAxis;

	YsListItem <YsDnmNode> *thisInTheList;

	void SetState(int sta);
	void SetShow(YSBOOL show);
	void InterpolateState(int staA,int staB,const double &param);
	void SetHeading(const double &h);
	void SetPitch(const double &p);
	void SetBank(const double &b);
};


class YsDnmPackedFile
{
public:
	char fn[256];
	YsArray <YsString> txt;

	void Initialize(void);
};

class YsDnm
{
protected:
	YsListContainer <YsDnmNode> nodeList;
	YsArray <YsDnmNode *> rootNode;

	static const char *const keyWordSource[];
	static class YsKeyWordList keyWordList;
public:
	static YsListAllocator <YsDnmNode> nodeAllocator;

	YsDnm();
	~YsDnm();
	void CleanUp(void);

	static YSBOOL IsPointVisible
	    (int np,const YsVec3 p[],
	     const YsMatrix4x4 &viewModelTfm,
		 const double &nearZ,const double &farZ,const double &tanFov);
	static YSBOOL IsBbxVisible
		    (const YsVec3 bbx[2],
		     const YsMatrix4x4 &viewModelTfm,
		     const double &nearZ,const double &farZ,const double &tanFov);


	YsDnmNode *CreateNode(void);
	void DeleteChildNodeNameTag(void);


public:
	YSRESULT LoadDnm(const char fn[]);
	YSRESULT LoadDnm(FILE *fp);
	YSRESULT LoadDnm(YsTextInputStream &inStream);

private:
	YSRESULT LoadDnm(YsTextInputStream &inStream,const char fn[]);
	YSRESULT BeginReadDnm(const char fn[]);
	YSRESULT ReadDnmOneLine(const char str[]);
	YSRESULT EndReadDnm(void);
public:

	YSRESULT SaveDnm(const char fn[]) const;
	YSRESULT SaveDnm(YsTextOutputStream &textOut) const;
protected:
	YSRESULT SaveDnmAddPackAll(YsTextOutputStream &textOut) const;
	YSRESULT SaveDnmAddPack(YsTextOutputStream &textOut,YsDnmNode *node) const;

protected:
	YSRESULT MakeConnectionAndRootNodeListByChildNameTag(void);
	YsDnmPackedFile *CreatePackedFile(const char fn[]);
	int readState; // -1:Error!!  0:Normal  1:Reading PLG  99:Loading Packed File  999:End of file
	int dnmVersion; // 0: Old dnm  1: Version 1
	int nPackedFileRemain,nVtxRemain;
	YsString readPath;
	int currentNodeType;
	YsDnmNode *currentNode;
	YsDnmPackedFile *currentPack;
	static YsListAllocator <YsDnmPackedFile> packedFileAllocator;
	YsListContainer <YsDnmPackedFile> packedFileList;

	void ComputeBoundingBox(void);
	void ComputeBoundingBox(YsBoundingBoxMaker3 &bbx,YsDnmNode *node,const YsMatrix4x4 &currentMat);
	YsVec3 bbx[2];

public:
	void Draw(const YsVec3 &pos,const YsAtt3 &att) const;
	void DrawWithFullKnowledge(const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,unsigned int drawFlag) const;
protected:
	void DrawNode(YsDnmNode *node,const YsMatrix4x4 &modelTfm) const;
	void DrawNodeWithFullKnowledge(YsDnmNode *node,const YsMatrix4x4 &viewModelTfm,unsigned int drawFlag) const;


public:
	void DrawVisual
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,
	     const double &nearZ,const double &farZ,const double &tanFov,unsigned int drawFlag) const;
	void DrawShadowVisual
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir,
	     const double &nearZ,const double &farZ,const double &tanFov) const;

public:
	void DrawSimpleShadow(const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir) const;
protected:
	void DrawNodeSimpleShadow(YsDnmNode *node,YsMatrix4x4 &current) const;


public:
	void ClearState(void);
	void SetState(int sta);
	void SetClassState(int clId,int sta);
	void InterpolateClassState(int clId,int staA,int staB,const double &param);
	void SetShow(int clId,YSBOOL show);

	void SetClassHeading(int clId,const double &h);
	void SetClassPitch(int clId,const double &p);
	void SetClassBank(int clId,const double &b);

	void SetPapiColor(const double &angle);
	void SetVasiColor(const double &angle);

	void GetBoundingBox(YsVec3 bbx[2]) const;


	int FindDnmNodeIdByTag(const char tag[]) const;
	void SetNodeHeadingPitch(int nodeId,const double &h,const double &p);
};

/* } */
#endif
