#include <ysclass.h>
#include "ysdnm.h"

#ifdef _WIN32
#include <windows.h>
#endif

#ifndef __APPLE__
	#include <GL/gl.h>
	#include <GL/glu.h>
	#include <GL/glext.h>
#elif defined(__APPLE_IOS__)
	#include <OpenGLES/ES2/gl.h>
	#include <OpenGLES/ES2/glext.h>
#else
	#include <OpenGL/gl.h>
	#include <OpenGL/glu.h>
	#include <OpenGL/glext.h>
#endif


#include <ysgl.h>


#if (defined(_WIN32) || defined(_WIN64)) && !defined(_CRT_SECURE_NO_WARNINGS)
	// This disables annoying warnings VC++ gives for use of C standard library.  Shut the mouth up.
	// This needs to come before header inclusion.
	#define _CRT_SECURE_NO_WARNINGS
#endif


extern YSBOOL ysScnGlUseRwLightTexture;
extern GLuint ysScnGlRwLightTex;


class YsDnmExternalVertexBufferLink
{
public:
	class Shaded
	{
	public:
		YsArray <GLfloat> vtx;
		YsArray <GLfloat> nom;
		YsArray <GLfloat> col;
	};
	class Unshaded
	{
	public:
		YsArray <GLfloat> vtx;
		YsArray <GLfloat> col;
	};
	class Light
	{
	public:
		YsArray <GLfloat> vtx;
		YsArray <GLfloat> col;
		YsArray <GLfloat> viewOffset;
	};


	Shaded solidShaded;
	Unshaded solidUnshaded;

	Shaded trspShaded;
	Unshaded trspUnshaded;

	mutable Light light;

public:
	YSBOOL IsEmpty(void) const;
	void CleanUp(void);
	void MakeShellVertexArray(const YsShellVisual &shl);
private:
	void AddShellTriangleToVertexArray(const YsShellVisual &shl,YsShellPolygonHandle plHd,const YsShellVertexHandle plVtHd[]);
};

YSBOOL YsDnmExternalVertexBufferLink::IsEmpty(void) const
{
	if(0==solidShaded.vtx.GetN() &&
	   0==solidUnshaded.vtx.GetN() &&
	   0==trspShaded.vtx.GetN() &&
	   0==trspUnshaded.vtx.GetN())
	{
		return YSTRUE;
	}
	return YSFALSE;
}

void YsDnmExternalVertexBufferLink::CleanUp(void)
{
	solidShaded.vtx.Clear();
	solidShaded.nom.Clear();
	solidShaded.col.Clear();

	solidUnshaded.vtx.Clear();
	solidUnshaded.col.Clear();

	trspShaded.vtx.Clear();
	trspShaded.nom.Clear();
	trspShaded.col.Clear();

	trspUnshaded.vtx.Clear();
	trspUnshaded.col.Clear();

	light.vtx.Clear();
	light.col.Clear();
	light.viewOffset.Clear();
}

void YsDnmExternalVertexBufferLink::MakeShellVertexArray(const YsShellVisual &shl)
{
	solidShaded.vtx.Clear();
	solidShaded.nom.Clear();
	solidShaded.col.Clear();

	solidUnshaded.vtx.Clear();
	solidUnshaded.col.Clear();

	trspShaded.vtx.Clear();
	trspShaded.nom.Clear();
	trspShaded.col.Clear();

	trspUnshaded.vtx.Clear();
	trspUnshaded.col.Clear();


	// Just set up buffer. Don't populate. >>
	light.vtx.Set(shl.light.GetN()*3*3,NULL);
	light.col.Set(shl.light.GetN()*4*3,NULL);
	light.viewOffset.Set(shl.light.GetN()*3*3,NULL);
	// Just set up buffer. Don't populate. <<


	shl.Encache();

	for(YsShellPolygonHandle plHd=NULL; NULL!=(plHd=shl.FindNextPolygon(plHd)); )
	{
		const int nPlVt=shl.GetNumVertexOfPolygon(plHd);;
		const YsShellVertexHandle *const plVtHd=shl.GetVertexListOfPolygon(plHd);;
		const unsigned int flags=shl.GetPolygon(plHd)->freeAttribute0;

		if(nPlVt==3)
		{
			AddShellTriangleToVertexArray(shl,plHd,plVtHd);
		}
		else if((flags&YsShellVisual::PLG_CONCAVE)==0)
		{
			for(int i=1; i<nPlVt-1; i++)
			{
				const YsShellVertexHandle triVtHd[3]=
				{
					plVtHd[0],
					plVtHd[i],
					plVtHd[i+1]
				};
				AddShellTriangleToVertexArray(shl,plHd,triVtHd);
			}
		}
		else
		{
			YsSword swd;
			YsArray <YsVec3> plVtPos;
			shl.GetVertexListOfPolygon(plVtPos,plHd);

			YsArray <int> plVtId;
			plVtId.Set(nPlVt,NULL);
			for(int i=0; i<nPlVt; i++)
			{
				plVtId[i]=shl.GetVertexIdFromHandle(plVtHd[i]);
			}


			swd.SetInitialPolygon(plVtPos.GetN(),plVtPos,plVtId);
			swd.Triangulate();
			for(int i=0; i<swd.GetNumPolygon(); i++)
			{
				const YsArray <int> *triVtId=swd.GetVertexIdList(i);
				if(triVtId->GetN()>=3)
				{
					const YsShellVertexHandle triVtHd[3]=
					{
						shl.GetVertexHandleFromId((*triVtId)[0]),
						shl.GetVertexHandleFromId((*triVtId)[1]),
						shl.GetVertexHandleFromId((*triVtId)[2])
					};
					AddShellTriangleToVertexArray(shl,plHd,triVtHd);
				}
			}
		}
	}
}

void YsDnmExternalVertexBufferLink::AddShellTriangleToVertexArray(const YsShellVisual &shl,YsShellPolygonHandle srcPlHd,const YsShellVertexHandle triVtHd[])
{
	YsVec3 defNom,vtxNom,tstNom;
	YsColor col;

	const unsigned int flags=shl.GetPolygon(srcPlHd)->freeAttribute0;
	shl.GetNormalOfPolygon(defNom,srcPlHd);

	shl.GetColorOfPolygon(col,srcPlHd);
	const YSBOOL lighting=((flags&YsShellVisual::PLG_NOSHADING)!=0 ? YSFALSE : YSTRUE);
	const unsigned int ahpla=(flags>>8)&255;

	const int alpha=255-ahpla;

	for(int i=0; i<3; i++)
	{
		YsVec3 nom;
		YsVec3 pos;

		shl.GetVertexPosition(pos,triVtHd[i]);
		const YsShellVertex *vtx=shl.GetVertex(triVtHd[i]);
		if((flags&YsShellVisual::PLG_TWOSIDED)==0 && vtx->freeAttribute0&YsShellVisual::VTX_ROUND)
		{
			nom=vtx->GetNormal();
		}
		else
		{
			nom=defNom;
		}

		if(lighting==YSTRUE)
		{
			if(ahpla==0)
			{
				solidShaded.vtx.Append((GLfloat)pos.x());
				solidShaded.vtx.Append((GLfloat)pos.y());
				solidShaded.vtx.Append((GLfloat)pos.z());

				solidShaded.nom.Append((GLfloat)nom.x());
				solidShaded.nom.Append((GLfloat)nom.y());
				solidShaded.nom.Append((GLfloat)nom.z());

				solidShaded.col.Append((GLfloat)col.Rd());
				solidShaded.col.Append((GLfloat)col.Gd());
				solidShaded.col.Append((GLfloat)col.Bd());
				solidShaded.col.Append((GLfloat)1.0);
			}
			else
			{
				trspShaded.vtx.Append((GLfloat)pos.x());
				trspShaded.vtx.Append((GLfloat)pos.y());
				trspShaded.vtx.Append((GLfloat)pos.z());

				trspShaded.nom.Append((GLfloat)nom.x());
				trspShaded.nom.Append((GLfloat)nom.y());
				trspShaded.nom.Append((GLfloat)nom.z());

				trspShaded.col.Append((GLfloat)col.Rd());
				trspShaded.col.Append((GLfloat)col.Gd());
				trspShaded.col.Append((GLfloat)col.Bd());
				trspShaded.col.Append(((GLfloat)alpha)/255.0F);
			}
		}
		else
		{
			if(ahpla==0)
			{
				solidUnshaded.vtx.Append((GLfloat)pos.x());
				solidUnshaded.vtx.Append((GLfloat)pos.y());
				solidUnshaded.vtx.Append((GLfloat)pos.z());

				solidUnshaded.col.Append((GLfloat)col.Rd());
				solidUnshaded.col.Append((GLfloat)col.Gd());
				solidUnshaded.col.Append((GLfloat)col.Bd());
				solidUnshaded.col.Append((GLfloat)1.0);
			}
			else
			{
				trspUnshaded.vtx.Append((GLfloat)pos.x());
				trspUnshaded.vtx.Append((GLfloat)pos.y());
				trspUnshaded.vtx.Append((GLfloat)pos.z());

				trspUnshaded.col.Append((GLfloat)col.Rd());
				trspUnshaded.col.Append((GLfloat)col.Gd());
				trspUnshaded.col.Append((GLfloat)col.Bd());
				trspUnshaded.col.Append(((GLfloat)alpha)/255.0F);
			}
		}
	}

	if(flags&YsShellVisual::PLG_TWOSIDED)
	{
		YsVec3 nom=-defNom;  // Two sided polygon : Always flat shading

		for(int i=2; i>=0; i--)
		{
			YsVec3 pos;
			shl.GetVertexPosition(pos,triVtHd[i]);

			if(lighting==YSTRUE)
			{
				if(ahpla==0)
				{
					solidShaded.vtx.Append((GLfloat)pos.x());
					solidShaded.vtx.Append((GLfloat)pos.y());
					solidShaded.vtx.Append((GLfloat)pos.z());

					solidShaded.nom.Append((GLfloat)nom.x());
					solidShaded.nom.Append((GLfloat)nom.y());
					solidShaded.nom.Append((GLfloat)nom.z());

					solidShaded.col.Append((GLfloat)col.Rd());
					solidShaded.col.Append((GLfloat)col.Gd());
					solidShaded.col.Append((GLfloat)col.Bd());
					solidShaded.col.Append((GLfloat)1.0);
				}
				else
				{
					trspShaded.vtx.Append((GLfloat)pos.x());
					trspShaded.vtx.Append((GLfloat)pos.y());
					trspShaded.vtx.Append((GLfloat)pos.z());

					trspShaded.nom.Append((GLfloat)nom.x());
					trspShaded.nom.Append((GLfloat)nom.y());
					trspShaded.nom.Append((GLfloat)nom.z());

					trspShaded.col.Append((GLfloat)col.Rd());
					trspShaded.col.Append((GLfloat)col.Gd());
					trspShaded.col.Append((GLfloat)col.Bd());
					trspShaded.col.Append(((GLfloat)alpha)/255.0F);
				}
			}
			else
			{
				if(ahpla==0)
				{
					solidUnshaded.vtx.Append((GLfloat)pos.x());
					solidUnshaded.vtx.Append((GLfloat)pos.y());
					solidUnshaded.vtx.Append((GLfloat)pos.z());

					solidUnshaded.col.Append((GLfloat)col.Rd());
					solidUnshaded.col.Append((GLfloat)col.Gd());
					solidUnshaded.col.Append((GLfloat)col.Bd());
					solidUnshaded.col.Append((GLfloat)1.0);
				}
				else
				{
					trspUnshaded.vtx.Append((GLfloat)pos.x());
					trspUnshaded.vtx.Append((GLfloat)pos.y());
					trspUnshaded.vtx.Append((GLfloat)pos.z());

					trspUnshaded.col.Append((GLfloat)col.Rd());
					trspUnshaded.col.Append((GLfloat)col.Gd());
					trspUnshaded.col.Append((GLfloat)col.Bd());
					trspUnshaded.col.Append(((GLfloat)alpha)/255.0F);
				}
			}
		}
	}
}

void FsTestPerformance(void)
{
}

static YSRESULT YsMakePlaneProjectionMatrix(YsMatrix4x4 &tfm,const YsVec3 &o,const YsVec3 &n,const YsVec3 &q)  // q:Direction of projection
{
	const double qn=q*n;
	const double on=o*n;

	if(qn!=0.0)
	{
		tfm.Set(1,1,1.0-(n.x()*q.x())/qn);
		tfm.Set(1,2,-n.y()*q.x()/qn);
		tfm.Set(1,3,-n.z()*q.x()/qn);
		tfm.Set(1,4,q.x()*on/qn);

		tfm.Set(2,1,-n.x()*q.y()/qn);
		tfm.Set(2,2,1.0-n.y()*q.y()/qn);
		tfm.Set(2,3,-n.z()*q.y()/qn);
		tfm.Set(2,4,q.y()*on/qn);

		tfm.Set(3,1,-n.x()*q.z()/qn);
		tfm.Set(3,2,-n.y()*q.z()/qn);
		tfm.Set(3,3,1.0-n.z()*q.z()/qn);
		tfm.Set(3,4,q.z()*on/qn);

		tfm.Set(4,1,0.0);
		tfm.Set(4,2,0.0);
		tfm.Set(4,3,0.0);
		tfm.Set(4,4,1.0);

		return YSOK;
	}

	return YSERR;
}

#ifndef CALLBACK
#define CALLBACK
#endif

static void DrawShellPlain(const YsShellVisual &shl)
{
	if(NULL!=shl.extBufLink)
	{
		if(YSTRUE==shl.extBufLink->IsEmpty())
		{
			shl.extBufLink->MakeShellVertexArray(shl);
		}

		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
			YsGLSLDrawPrimitiveVtxNomColfv(
			    YsGLSLSharedVariColorShaded3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->solidShaded.vtx.GetN()/3,
			    shl.extBufLink->solidShaded.vtx,
			    shl.extBufLink->solidShaded.nom,
			    shl.extBufLink->solidShaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
			YsGLSLDrawPrimitiveVtxColfv(
			    YsGLSLSharedVariColor3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->solidUnshaded.vtx.GetN()/3,
			    shl.extBufLink->solidUnshaded.vtx,
			    shl.extBufLink->solidUnshaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
			YsGLSLDrawPrimitiveVtxNomColfv(
			    YsGLSLSharedVariColorShaded3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->trspShaded.vtx.GetN()/3,
			    shl.extBufLink->trspShaded.vtx,
			    shl.extBufLink->trspShaded.nom,
			    shl.extBufLink->trspShaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer())
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
			YsGLSLDrawPrimitiveVtxColfv(
			    YsGLSLSharedVariColor3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->trspUnshaded.vtx.GetN()/3,
			    shl.extBufLink->trspUnshaded.vtx,
			    shl.extBufLink->trspUnshaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		}
	}
}

static void DrawShellFlatColor(const YsShellVisual &shl)
{
	if(NULL!=shl.extBufLink)
	{
		if(YSTRUE==shl.extBufLink->IsEmpty())
		{
			shl.extBufLink->MakeShellVertexArray(shl);
		}

		YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();

		// YsGLSLSharedFlat3DRenderer must be activated before this function.

		if(NULL!=renderer)
		{
			YsGLSLDrawPrimitiveVtxfv(
			    renderer,
			    GL_TRIANGLES,
			    shl.extBufLink->solidShaded.vtx.GetN()/3,
			    shl.extBufLink->solidShaded.vtx);

			YsGLSLDrawPrimitiveVtxfv(
			    renderer,
			    GL_TRIANGLES,
			    shl.extBufLink->solidUnshaded.vtx.GetN()/3,
			    shl.extBufLink->solidUnshaded.vtx);

			YsGLSLDrawPrimitiveVtxfv(
			    renderer,
			    GL_TRIANGLES,
			    shl.extBufLink->trspShaded.vtx.GetN()/3,
			    shl.extBufLink->trspShaded.vtx);

			YsGLSLDrawPrimitiveVtxfv(
			    renderer,
			    GL_TRIANGLES,
			    shl.extBufLink->trspUnshaded.vtx.GetN()/3,
			    shl.extBufLink->trspUnshaded.vtx);
		}
	}
}

static void DrawShellWithFullKnowledge(const YsShellVisual &shl,const YsMatrix4x4 &viewModelTfm ,unsigned int drawFlag)
{
	if(NULL!=shl.extBufLink)
	{
		if(YSTRUE==shl.extBufLink->IsEmpty())
		{
			shl.extBufLink->MakeShellVertexArray(shl);
		}

		GLfloat viewModelMat[16];
		viewModelTfm.GetOpenGlCompatibleMatrix(viewModelMat);

		YsGLInvertZfv(viewModelMat);  // From left-hand coord to right-hand coord.

		YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),viewModelMat);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

		YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),viewModelMat);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());


		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer() && 0!=(drawFlag&YSDNM_DRAWOPAQUE))
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
			YsGLSLDrawPrimitiveVtxNomColfv(
			    YsGLSLSharedVariColorShaded3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->solidShaded.vtx.GetN()/3,
			    shl.extBufLink->solidShaded.vtx,
			    shl.extBufLink->solidShaded.nom,
			    shl.extBufLink->solidShaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer() && 0!=(drawFlag&YSDNM_DRAWOPAQUE))
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
			YsGLSLDrawPrimitiveVtxColfv(
			    YsGLSLSharedVariColor3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->solidUnshaded.vtx.GetN()/3,
			    shl.extBufLink->solidUnshaded.vtx,
			    shl.extBufLink->solidUnshaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColorShaded3DRenderer() && 0!=(drawFlag&YSDNM_DRAWTRANSPARENT))
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
			YsGLSLDrawPrimitiveVtxNomColfv(
			    YsGLSLSharedVariColorShaded3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->trspShaded.vtx.GetN()/3,
			    shl.extBufLink->trspShaded.vtx,
			    shl.extBufLink->trspShaded.nom,
			    shl.extBufLink->trspShaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		}
		if(NULL!=YsGLSLSharedVariColor3DRenderer() && 0!=(drawFlag&YSDNM_DRAWTRANSPARENT))
		{
			YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
			YsGLSLDrawPrimitiveVtxColfv(
			    YsGLSLSharedVariColor3DRenderer(),
			    GL_TRIANGLES,
			    shl.extBufLink->trspUnshaded.vtx.GetN()/3,
			    shl.extBufLink->trspUnshaded.vtx,
			    shl.extBufLink->trspUnshaded.col);
			YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		}
	}



	if(shl.flags&YsShellVisual::SHL_DRAWPOINT)
	{
/*		YSBOOL first;

		plHd=NULL;
		first=YSTRUE;
		glDisable(GL_LIGHTING);
		glBegin(GL_POINTS);
		while((plHd=shl.FindNextPolygon(plHd))!=NULL)
		{
			YsColor col;
			YsVec3 pos;

			shl.GetColorOfPolygon(col,plHd);
			shl.GetCenterOfPolygon(pos,plHd);

			if(first==YSTRUE || col!=prevColor)
			{
				glColor3d(col.Rd(),col.Gd(),col.Bd());
				prevColor=col;
			}

			glVertex3dv(pos);

			first=YSFALSE;
		}
		glEnd(); */
	}
}

void YsShellVisual::AllocateD3dExternalVertexBufferLink(void)
{
	extBufLink=new YsDnmExternalVertexBufferLink;
}

void YsShellVisual::FreeD3dExternalVertexBufferLink(void)
{
	delete extBufLink;
}

void YsShellVisual::ReleaseD3dExternalVertexBuffer(void)
{
	extBufLink->CleanUp();
}

void YsShellVisual::Draw(void) const
{
	glEnable(GL_CULL_FACE);
	DrawShellPlain(*this);
}

void YsShellVisual::Draw(const YsMatrix4x4 &viewModelTfm) const
{
	DrawShellWithFullKnowledge(*this,viewModelTfm,YSDNM_DRAWALL);
}

void YsShellVisual::Draw(const YsVec3 &pos,const YsAtt3 &att) const
{
	GLfloat prevModelViewMat[16];

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLGet3DRendererModelViewfv(prevModelViewMat,YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsMatrix4x4 tfm(YSFALSE);
	tfm.CreateFromOpenGlCompatibleMatrix(prevModelViewMat);

	tfm.Translate(pos);
	tfm.RotateXZ(att.h());
	tfm.RotateZY(att.p());
	tfm.RotateXY(att.b());

	GLfloat newModelViewMat[16];
	tfm.GetOpenGlCompatibleMatrix(newModelViewMat);

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),newModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),newModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());

	Draw();

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),prevModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),prevModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
}

void YsShellVisual::DrawVisual
	    (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,
	     const double &nearZ,const double &farZ,const double &tanFov,unsigned int drawFlag) const
{
	YsMatrix4x4 viewModelTfm;
	viewModelTfm=viewTfm;
	viewModelTfm.Translate(pos);
	viewModelTfm.RotateXZ(att.h());
	viewModelTfm.RotateZY(att.p());
	viewModelTfm.RotateXY(att.b());

	YsVec3 bbx[2];
	GetBoundingBox(bbx[0],bbx[1]);

	if(YsDnm::IsBbxVisible(bbx,viewModelTfm,nearZ,farZ,tanFov)==YSTRUE)
	{
		GLfloat prevModelViewMat[16];
		YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		YsGLSLGet3DRendererModelViewfv(prevModelViewMat,YsGLSLSharedVariColorShaded3DRenderer());
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());


		GLfloat newModelViewMat[16];
		viewModelTfm.GetOpenGlCompatibleMatrix(newModelViewMat);

		YsGLInvertZfv(newModelViewMat);  // From left-hand coord to right-hand coord.

		YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),newModelViewMat);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

		YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),newModelViewMat);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());

		DrawVisual(viewModelTfm,drawFlag);

		YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),prevModelViewMat);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

		YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),prevModelViewMat);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	}
}

void YsShellVisual::DrawVisual(const YsMatrix4x4 &viewModelTfm,unsigned int drawFlag) const
{
	DrawLightVisual(viewModelTfm);
	DrawShellWithFullKnowledge(*this,viewModelTfm,drawFlag);
}

void YsShellVisual::DrawLightVisual(const YsMatrix4x4 &viewModelTfm) const
{
	if(NULL!=extBufLink)
	{
		const int nLightMax=extBufLink->light.vtx.GetN()/9;
		int nLightToDraw=0;

		for(int i=0; i<light.GetN() && nLightToDraw<nLightMax; i++)
		{
			const YsVec3 &cen=light[i].cen;
			const YsVec3 &nom=light[i].nom;
			const YsColor &col=light[i].col;

			if(nom!=YsOrigin())
			{
				YsVec3 nomTfm,cenTfm;
				viewModelTfm.Mul(nomTfm,nom,0.0);
				viewModelTfm.Mul(cenTfm,cen,1.0);
				if(nomTfm*cenTfm>0.0)
				{
					continue;
				}
			}

			extBufLink->light.vtx[nLightToDraw*9   ]=(GLfloat)cen.x();
			extBufLink->light.vtx[nLightToDraw*9+ 1]=(GLfloat)cen.y();
			extBufLink->light.vtx[nLightToDraw*9+ 2]=(GLfloat)cen.z();

			extBufLink->light.col[nLightToDraw*12   ]=(GLfloat)col.Rd();
			extBufLink->light.col[nLightToDraw*12+ 1]=(GLfloat)col.Gd();
			extBufLink->light.col[nLightToDraw*12+ 2]=(GLfloat)col.Bd();
			extBufLink->light.col[nLightToDraw*12+ 3]=1.0f;

			extBufLink->light.viewOffset[nLightToDraw*9   ]=1.7320508f;  // <- Sensitive to culling.
			extBufLink->light.viewOffset[nLightToDraw*9+ 1]=-1.0f;
			extBufLink->light.viewOffset[nLightToDraw*9+ 2]=0.1f;


			extBufLink->light.vtx[nLightToDraw*9+ 3]=(GLfloat)cen.x();
			extBufLink->light.vtx[nLightToDraw*9+ 4]=(GLfloat)cen.y();
			extBufLink->light.vtx[nLightToDraw*9+ 5]=(GLfloat)cen.z();

			extBufLink->light.col[nLightToDraw*12+ 4]=(GLfloat)col.Rd();
			extBufLink->light.col[nLightToDraw*12+ 5]=(GLfloat)col.Gd();
			extBufLink->light.col[nLightToDraw*12+ 6]=(GLfloat)col.Bd();
			extBufLink->light.col[nLightToDraw*12+ 7]=1.0f;

			extBufLink->light.viewOffset[nLightToDraw*9+ 3]=0.0f;
			extBufLink->light.viewOffset[nLightToDraw*9+ 4]=2.0f;
			extBufLink->light.viewOffset[nLightToDraw*9+ 5]=0.1f;


			extBufLink->light.vtx[nLightToDraw*9+ 6]=(GLfloat)cen.x();
			extBufLink->light.vtx[nLightToDraw*9+ 7]=(GLfloat)cen.y();
			extBufLink->light.vtx[nLightToDraw*9+ 8]=(GLfloat)cen.z();

			extBufLink->light.col[nLightToDraw*12+ 8]=(GLfloat)col.Rd();
			extBufLink->light.col[nLightToDraw*12+ 9]=(GLfloat)col.Gd();
			extBufLink->light.col[nLightToDraw*12+10]=(GLfloat)col.Bd();
			extBufLink->light.col[nLightToDraw*12+11]=1.0f;

			extBufLink->light.viewOffset[nLightToDraw*9+ 6]=-1.7320508f;
			extBufLink->light.viewOffset[nLightToDraw*9+ 7]=-1.0f;
			extBufLink->light.viewOffset[nLightToDraw*9+ 8]=0.1f;

			++nLightToDraw;
		}

		if(0<nLightToDraw)
		{
			GLfloat modelViewMat[16];
			viewModelTfm.GetOpenGlCompatibleMatrix(modelViewMat);
			YsGLInvertZfv(modelViewMat);

			YsGLSLUse3DRenderer(YsGLSLSharedFlash3DRenderer());
			YsGLSLSet3DRendererModelViewfv(YsGLSLSharedFlash3DRenderer(),modelViewMat);
			YsGLSLDrawPrimitiveVtxColVoffsetfv(YsGLSLSharedFlash3DRenderer(),nLightToDraw*3,extBufLink->light.vtx,extBufLink->light.col,extBufLink->light.viewOffset);
			YsGLSLEndUse3DRenderer(YsGLSLSharedFlash3DRenderer());
		}
	}
}

void YsShellVisual::DrawForMakingOpenGlDisplayList(unsigned int /*listId*/ )
{
	if(NULL!=extBufLink && YSTRUE==extBufLink->IsEmpty())
	{
		extBufLink->MakeShellVertexArray(*this);
	}
}

void YsShellVisual::DrawVisualLowPriorityMode(
    const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,
	const double &nearZ,const double &farZ,const double &tanFov) const
{
	DrawVisual(viewTfm,pos,att,nearZ,farZ,tanFov,YSDNM_DRAWALL);
}

void YsShellVisual::DrawSimpleShadow(const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir) const
{
	YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsGLSLUse3DRenderer(renderer);


	GLfloat prevModelViewMat[16];
	YsGLSLGet3DRendererModelViewfv(prevModelViewMat,renderer);


	YsMatrix4x4 shadowCastMat(YSFALSE);
	YsMakePlaneProjectionMatrix(shadowCastMat,terOrg,terNom,lightDir);



	const GLfloat black[4]={0,0,0,1};
	YsGLSLSet3DRendererUniformColorfv(renderer,black);


	YsMatrix4x4 current(YSFALSE);
	current.CreateFromOpenGlCompatibleMatrix(prevModelViewMat);
	current*=shadowCastMat;
	current.Translate(pos);
	current.RotateXZ(att.h());
	current.RotateZY(att.p());
	current.RotateXY(att.b());


	GLfloat modelViewMat[16];
	current.GetOpenGlCompatibleMatrix(modelViewMat);
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedFlat3DRenderer(),modelViewMat);

	DrawShellFlatColor(*this);


	YsGLSLSet3DRendererModelViewfv(renderer,prevModelViewMat);

	YsGLSLEndUse3DRenderer(renderer);
}


////////////////////////////////////////////////////////////


void YsDnm::Draw(const YsVec3 &pos,const YsAtt3 &att) const
{
	GLfloat prevTfm[16];
	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLGet3DRendererModelViewfv(prevTfm,YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());


	GLfloat tfm[16];
	YsGLCopyMatrixfv(tfm,prevTfm);
	YsGLMultMatrixTranslationfv(tfm,(GLfloat)pos.x(),(GLfloat)pos.y(),(GLfloat)pos.z());

	YsGLMultMatrixRotationXZfv(tfm,(GLfloat)att.h());
	YsGLMultMatrixRotationZYfv(tfm,(GLfloat)att.p());
	YsGLMultMatrixRotationXYfv(tfm,(GLfloat)att.b());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),tfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),tfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());


	YsMatrix4x4 current;

	current.Translate(pos);
	current.RotateXZ(att.h());
	current.RotateZY(att.p());
	current.RotateXY(att.b());

	for(int i=0; i<rootNode.GetN(); i++)
	{
		DrawNode(rootNode[i],current);
	}

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),prevTfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),prevTfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
}

void YsDnm::DrawNode(YsDnmNode *node,const YsMatrix4x4 &current) const
{
	GLfloat prevTfm[16];
	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLGet3DRendererModelViewfv(prevTfm,YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());


	YsMatrix4x4 newCurrent(current);

	newCurrent.Translate(node->posInParent);

	newCurrent.RotateXZ(node->attInParent.h());
	newCurrent.RotateZY(node->attInParent.p());
	newCurrent.RotateXY(node->attInParent.b());

	newCurrent.Translate(node->currentState.relPos);

	newCurrent.RotateXZ(node->currentState.relAtt.h());
	newCurrent.RotateZY(node->currentState.relAtt.p());
	newCurrent.RotateXY(node->currentState.relAtt.b());

	newCurrent.Translate(-node->localRotationCenter);


	GLfloat tfm[16];
	YsGLCopyMatrixfv(tfm,prevTfm);
	YsGLMultMatrixTranslationfv(tfm,(GLfloat)node->posInParent.x(),(GLfloat)node->posInParent.y(),(GLfloat)node->posInParent.z());

	YsGLMultMatrixRotationXZfv(tfm,(GLfloat)node->attInParent.h());
	YsGLMultMatrixRotationZYfv(tfm,(GLfloat)node->attInParent.p());
	YsGLMultMatrixRotationXYfv(tfm,(GLfloat)node->attInParent.b());

	YsGLMultMatrixTranslationfv(tfm,(GLfloat)node->currentState.relPos.x(),(GLfloat)node->currentState.relPos.y(),(GLfloat)node->currentState.relPos.z());

	YsGLMultMatrixRotationXZfv(tfm,(GLfloat)node->currentState.relAtt.h());
	YsGLMultMatrixRotationZYfv(tfm,(GLfloat)node->currentState.relAtt.p());
	YsGLMultMatrixRotationXYfv(tfm,(GLfloat)node->currentState.relAtt.b());

	YsGLMultMatrixTranslationfv(tfm,-(GLfloat)node->localRotationCenter.x(),-(GLfloat)node->localRotationCenter.y(),-(GLfloat)node->localRotationCenter.z());


	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),tfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),tfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());


	if(node->currentState.show==YSTRUE)
	{
		node->Draw();
	}

	for(int i=0; i<node->children.GetN(); i++)
	{
		DrawNode(node->children[i],newCurrent);
	}

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),prevTfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),prevTfm);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
}

void YsDnm::DrawWithFullKnowledge
   (const YsMatrix4x4 &viewTfm,const YsVec3 &pos,const YsAtt3 &att,unsigned int drawFlag) const
{
	GLfloat prevModelViewMat[16];
	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLGet3DRendererModelViewfv(prevModelViewMat,YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsMatrix4x4 viewModelTfm=viewTfm;

	viewModelTfm.Translate(pos);
	viewModelTfm.RotateXZ(att.h());
	viewModelTfm.RotateZY(att.p());
	viewModelTfm.RotateXY(att.b());

	for(int i=0; i<rootNode.GetN(); i++)
	{
		DrawNodeWithFullKnowledge(rootNode[i],viewModelTfm,drawFlag);
	}

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),prevModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),prevModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
}

void YsDnm::DrawNodeWithFullKnowledge(YsDnmNode *node,const YsMatrix4x4 &viewModelTfm,unsigned int drawFlag) const
{
	glEnable(GL_CULL_FACE);

	GLfloat prevModelViewMat[16];
	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLGet3DRendererModelViewfv(prevModelViewMat,YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsMatrix4x4 newCurrent=viewModelTfm;

	newCurrent.Translate(node->posInParent);

	newCurrent.RotateXZ(node->attInParent.h());
	newCurrent.RotateZY(node->attInParent.p());
	newCurrent.RotateXY(node->attInParent.b());

	newCurrent.Translate(node->currentState.relPos);

	newCurrent.RotateXZ(node->currentState.relAtt.h());
	newCurrent.RotateZY(node->currentState.relAtt.p());
	newCurrent.RotateXY(node->currentState.relAtt.b());

	newCurrent.Translate(-node->localRotationCenter);

	if(node->currentState.show==YSTRUE)
	{
		node->DrawVisual(newCurrent,drawFlag); // DrawShellWithFullKnowledge(*node,newCurrent,drawFlag);
	}

	for(int i=0; i<node->children.GetN(); i++)
	{
		DrawNodeWithFullKnowledge(node->children[i],newCurrent,drawFlag);
	}

	YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),prevModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());

	YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),prevModelViewMat);
	YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
}

void YsDnm::DrawSimpleShadow(const YsVec3 &pos,const YsAtt3 &att,const YsVec3 &terOrg,const YsVec3 &terNom,const YsVec3 &lightDir) const
{
	YsGLSL3DRenderer *renderer=YsGLSLSharedFlat3DRenderer();
	YsGLSLUse3DRenderer(renderer);


	GLfloat prevModelViewMat[16];
	YsGLSLGet3DRendererModelViewfv(prevModelViewMat,renderer);


	YsMatrix4x4 shadowCastMat(YSFALSE);
	YsMakePlaneProjectionMatrix(shadowCastMat,terOrg,terNom,lightDir);



	const GLfloat black[4]={0,0,0,1};
	YsGLSLSet3DRendererUniformColorfv(renderer,black);


	YsMatrix4x4 current(YSFALSE);
	current.CreateFromOpenGlCompatibleMatrix(prevModelViewMat);
	current*=shadowCastMat;
	current.Translate(pos);
	current.RotateXZ(att.h());
	current.RotateZY(att.p());
	current.RotateXY(att.b());

	for(int i=0; i<rootNode.GetN(); i++)
	{
		DrawNodeSimpleShadow(rootNode[i],current);
	}



	YsGLSLSet3DRendererModelViewfv(renderer,prevModelViewMat);

	YsGLSLEndUse3DRenderer(renderer);
}

void YsDnm::DrawNodeSimpleShadow(YsDnmNode *node,YsMatrix4x4 &current) const
{
	YsMatrix4x4 newCurrent=current;

	newCurrent.Translate(node->posInParent);

	newCurrent.RotateXZ(node->attInParent.h());
	newCurrent.RotateZY(node->attInParent.p());
	newCurrent.RotateXY(node->attInParent.b());

	newCurrent.Translate(node->currentState.relPos);

	newCurrent.RotateXZ(node->currentState.relAtt.h());
	newCurrent.RotateZY(node->currentState.relAtt.p());
	newCurrent.RotateXY(node->currentState.relAtt.b());

	newCurrent.Translate(-node->localRotationCenter);


	GLfloat modelViewMat[16];
	newCurrent.GetOpenGlCompatibleMatrix(modelViewMat);
	YsGLSLSet3DRendererModelViewfv(YsGLSLSharedFlat3DRenderer(),modelViewMat);

	if(node->currentState.show==YSTRUE)
	{
		DrawShellFlatColor(*node);
	}

	for(int i=0; i<node->children.GetN(); i++)
	{
		DrawNodeSimpleShadow(node->children[i],newCurrent);
	}
}


