#include "ysbitmaparray.h"


YsBitmapArray::YsBitmapArray()
{
	bmpArray=NULL;
	CleanUp();
}

YsBitmapArray::YsBitmapArray(const YsBitmapArray &from)
{
	bmpArray=NULL;
	CopyFrom(from);
}

const YsBitmapArray &YsBitmapArray::operator=(const YsBitmapArray &from)
{
	CopyFrom(from);
	return *this;
}

void YsBitmapArray::CopyFrom(const YsBitmapArray &from)
{
	CleanUp();

	totalWid=from.totalWid;
	totalHei=from.totalHei;
	tileWid=from.tileWid;
	tileHei=from.tileHei;
	nDivX=from.nDivX;
	nDivY=from.nDivY;

	if(0<nDivX*nDivY)
	{
		bmpArray=new YsBitmap [nDivX*nDivY];
		for(int i=0; i<nDivX*nDivY; ++i)
		{
			bmpArray[i]=from.bmpArray[i];
		}
	}
}

YsBitmapArray::~YsBitmapArray()
{
	ResizeArray(0);
}

void YsBitmapArray::CleanUp(void)
{
	totalWid=0;
	totalHei=0;
	tileWid=0;
	tileHei=0;
	nDivX=0;
	nDivY=0;
	if(NULL!=bmpArray)
	{
		delete [] bmpArray;
		bmpArray=NULL;
	}
}

YSRESULT YsBitmapArray::SetAndBreakUp(const YsBitmap &from,int maxBlockWidth,int maxBlockHeight)
{
	CleanUp();
	totalWid=from.GetWidth();
	totalHei=from.GetHeight();

	int blockWid=1,blockHei=1;
	while(blockWid*2<=maxBlockWidth)
	{
		blockWid*=2;
	}
	while(blockHei*2<=maxBlockHeight)
	{
		blockHei*=2;
	}

	totalWid=from.GetWidth();
	totalHei=from.GetHeight();
	nDivX=(totalWid+blockWid-1)/blockWid;
	nDivY=(totalHei+blockHei-1)/blockHei;

	if(0==nDivX || 0==nDivY)
	{
		nDivX=0;
		nDivY=0;
		return YSERR;
	}

	tileWid=blockWid;
	tileHei=blockHei;

	ResizeArray(nDivX*nDivY);
	for(unsigned int y=0; y<nDivY; ++y)
	{
		const unsigned int fromY0=y*blockHei;
		int tileHei;
		if(blockHei<=totalHei-fromY0)
		{
			tileHei=blockHei;
		}
		else
		{
			tileHei=1;
			while(tileHei<totalHei-fromY0)
			{
				tileHei*=2;
			}
		}

		for(unsigned int x=0; x<nDivX; ++x)
		{
			const unsigned int fromX0=x*blockWid;

			int offset=y*nDivX+x;

			int tileWid;
			if(blockWid<=totalWid-fromX0)
			{
				tileWid=blockWid;
			}
			else
			{
				tileWid=1;
				while(tileWid<totalWid-fromX0)
				{
					tileWid*=2;
				}
			}

			YsBitmap &tile=bmpArray[offset];
			tile.PrepareBitmap(tileWid,tileHei);
			tile.Clear(0,0,0,0);

			tile.Copy(from,fromX0,fromY0,0,0,tileWid,tileHei);
		}
	}

	return YSOK;
}

void YsBitmapArray::ResizeArray(int n)
{
	if(NULL!=bmpArray)
	{
		delete [] bmpArray;
		bmpArray=NULL;
	}
	if(0<n)
	{
		bmpArray=new YsBitmap [n];
	}
}

unsigned int YsBitmapArray::GetTotalWidth(void) const
{
	return totalWid;
}

unsigned int YsBitmapArray::GetTotalHeight(void) const
{
	return totalHei;
}

unsigned int YsBitmapArray::GetNumBlockX(void) const
{
	return nDivX;
}

unsigned int YsBitmapArray::GetNumBlockY(void) const
{
	return nDivY;
}

unsigned int YsBitmapArray::GetTileWidth(void) const
{
	return tileWid;
}

unsigned int YsBitmapArray::GetTileHeight(void) const
{
	return tileHei;
}

const YsBitmap *YsBitmapArray::GetTileBitmap(unsigned int x,unsigned int y) const
{
	if(0<=x && x<GetNumBlockX() && 0<=y && y<GetNumBlockY())
	{
		return &bmpArray[y*GetNumBlockX()+x];
	}
	return NULL;
}

YsBitmap *YsBitmapArray::GetTileBitmap(unsigned int x,unsigned int y)
{
	if(0<=x && x<GetNumBlockX() && 0<=y && y<GetNumBlockY())
	{
		return &bmpArray[y*GetNumBlockX()+x];
	}
	return NULL;
}
