#ifndef YSARRAYFOREDIT_IS_INCLUDED
#define YSARRAYFOREDIT_IS_INCLUDED
/* { */

#include "ysarray.h"
#include "yssegarray.h"


template <class T,const int bitShift>
class YsEditArray;


/*!  This is an object handle for YsEditArray.
*/
template <class T,const int bitShift=8>
class YsEditArrayObjectHandle
{
	friend class YsEditArray <T,bitShift>;

private:
	enum
	{
		NULLINDEX=0x7fffffff
	};

	YSSIZE_T index;

public:
	YsEditArrayObjectHandle();
	YsEditArrayObjectHandle(void *ptr);
	YsEditArrayObjectHandle(const YsEditArrayObjectHandle <T,bitShift> &incoming);

	inline const YsEditArrayObjectHandle &CopyFrom(void *ptr);
	inline const YsEditArrayObjectHandle &CopyFrom(const YsEditArrayObjectHandle <T,bitShift> &incoming);
	inline int IsEqual(void *ptr) const;
	inline int IsNotEqual(void *ptr) const;
	inline int IsEqual(const YsEditArrayObjectHandle <T,bitShift> &incoming) const;
	inline int IsNotEqual(const YsEditArrayObjectHandle <T,bitShift> &incoming) const;

	inline const YsEditArrayObjectHandle &operator=(void *ptr);
	inline int operator==(const YsEditArrayObjectHandle <T,bitShift> &incoming);
	inline int operator!=(const YsEditArrayObjectHandle <T,bitShift> &incoming);
	inline int operator==(void *ptr);
	inline int operator!=(void *ptr);
};

template <class T,const int bitShift>
YsEditArrayObjectHandle <T,bitShift>::YsEditArrayObjectHandle()
{
	index=NULLINDEX;
}

template <class T,const int bitShift>
YsEditArrayObjectHandle <T,bitShift>::YsEditArrayObjectHandle(void *ptr)
{
	CopyFrom(ptr);
}

template <class T,const int bitShift>
YsEditArrayObjectHandle <T,bitShift>::YsEditArrayObjectHandle(const YsEditArrayObjectHandle <T,bitShift> &incoming)
{
	CopyFrom(incoming);
}

template <class T,const int bitShift>
inline const YsEditArrayObjectHandle <T,bitShift> &YsEditArrayObjectHandle <T,bitShift>::CopyFrom(void *ptr)
{
	if(NULL==ptr)
	{
		this->index=NULLINDEX;
	}
	else
	{
		YsPrintf("%s\n",__FUNCTION__);
		YsPrintf("Only NULL pointer can be directly copied to YsEditArrayObjectHandle\n");
	}
	return *this;
}

template <class T,const int bitShift>
inline const YsEditArrayObjectHandle <T,bitShift> &YsEditArrayObjectHandle <T,bitShift>::CopyFrom(const YsEditArrayObjectHandle <T,bitShift> &incoming)
{
	this->index=incoming.index;
	return *this;
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::IsEqual(void *ptr) const
{
	if(NULL==ptr)
	{
		return NULLINDEX==this->index;
	}
	YsPrintf("%s\n",__FUNCTION__);
	YsPrintf("Only NULL pointer can be directly compared with YsEditArrayObjectHandle\n");
	return 0;
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::IsNotEqual(void *ptr) const
{
	if(NULL==ptr)
	{
		return NULLINDEX!=this->index;
	}
	YsPrintf("%s\n",__FUNCTION__);
	YsPrintf("Only NULL pointer can be directly compared with YsEditArrayObjectHandle\n");
	return 0;
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::IsEqual(const YsEditArrayObjectHandle <T,bitShift> &incoming) const
{
	return this->index==incoming.index;
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::IsNotEqual(const YsEditArrayObjectHandle <T,bitShift> &incoming) const
{
	return this->index!=incoming.index;
}

template <class T,const int bitShift>
inline const YsEditArrayObjectHandle <T,bitShift> &YsEditArrayObjectHandle <T,bitShift>::operator=(void *ptr)
{
	return CopyFrom(ptr);
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::operator==(const YsEditArrayObjectHandle <T,bitShift> &incoming)
{
	return IsEqual(incoming);
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::operator!=(const YsEditArrayObjectHandle <T,bitShift> &incoming)
{
	return IsNotEqual(incoming);
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::operator==(void *ptr)
{
	return IsEqual(ptr);
}

template <class T,const int bitShift>
inline int YsEditArrayObjectHandle <T,bitShift>::operator!=(void *ptr)
{
	return IsNotEqual(ptr);
}

template <class T,const int bitShift>
inline int operator==(void *ptr,const YsEditArrayObjectHandle <T,bitShift> &obHd)
{
	return obHd.IsEqual(ptr);
}

template <class T,const int bitShift>
inline int operator!=(void *ptr,const YsEditArrayObjectHandle <T,bitShift> &obHd)
{
	return obHd.IsNotEqual(ptr);
}

////////////////////////////////////////////////////////////

/*!  This is a template array class for editing.  Uses YsSegmentedArray as a background data structure.
*/
template <class T,const int bitShift=8>
class YsEditArray
{
public:
	/*! The object in the YsEditArray can be one of ALIVE, DELETED, or FROZEN.
	*/
	enum
	{
		ALIVE=0,
		DELETED=1,
		FROZEN=2
	};

private:
	// Actually I want to add Lock/Unlock

	unsigned int localSearchKeySeed;
	unsigned int *searchKeySeed;
	class ArrayObject
	{
	public:
		unsigned int attrib;  // bit0-1 STATE(ALIVE,DELETED,FROZEN)  bit 2-31 search key
		T dat;

		inline void SetSearchKey(unsigned int key)
		{
			attrib&=3;
			attrib|=(key<<2);
		}
		inline unsigned int GetSearchKey(void) const
		{
			return (attrib>>2);
		}
		inline void SetState(unsigned int state)
		{
			attrib&=~(unsigned int)3;
			attrib|=(state&3);
		}
		inline unsigned int GetState(void) const
		{
			return (attrib&3);
		}
	};

	YsSegmentedArray <ArrayObject,bitShift> objArray;
	YsArray <YSSIZE_T> deletedIndex; // <- Ready for recycling
	YSSIZE_T nFrozen;

	YSBOOL searchEnabled;
	YsHashTable <YSSIZE_T> searchKeyToIndex;

public:
	/*! Default constructor
	*/
	YsEditArray();

	/*! Clears the array. */
	void CleanUp(void);

	/*! Make this object to use a shared external search-key seed.
	*/
	void UseSharedSearchKeySeed(unsigned int &seedPtr);

	/*! Enable search table.  When the search is enabled, an array object can be
	    seached by FindObject or FindFrozenObject function.
	*/
	void EnableSearch(void);

	/*! Disable search table.
	*/
	void DisableSearch(void);

	/*! Returns number of objects ALIVE in the array.
	*/
	YSSIZE_T GetN(void) const;

	/*! Returns YSTRUE if the object is frozen (or tentatively deleted), or YSFALSE otherwise. */
	YSBOOL IsFrozen(YsEditArrayObjectHandle<T,bitShift> hd) const;

	/*! Create a new object and returns the handle.
	*/
	YsEditArrayObjectHandle<T,bitShift> Create(void);

	/*! Access the object. */
	T *operator[] (YsEditArrayObjectHandle<T,bitShift> hd);

	/*! Access the object. */
	const T *operator[] (YsEditArrayObjectHandle<T,bitShift> hd) const;

	/*! Access the object that is tentatively deleted. */
	const T *GetFrozenObject(YsEditArrayObjectHandle<T,bitShift> hd) const;

	/*! Appends an object.  T must be safe for the copy operator. */
	YsEditArrayObjectHandle<T,bitShift> Append(const T &incoming);

	/*! Delete an object.  The state of the object pointed by the handle must be ALIVE.
	*/
	YSRESULT Delete(YsEditArrayObjectHandle<T,bitShift> hd);

	/*! Freeze an object.  Frozen object is temporarily deleted and can later be melted (or undeleted.)
	*/
	YSRESULT Freeze(YsEditArrayObjectHandle<T,bitShift> hd);

	/*! Melt (undelete) an object that was temporarily deleted by Freeze function.
	*/
	YSRESULT Melt(YsEditArrayObjectHandle<T,bitShift> hd);

	/*! Delete a frozen object.  The state of the object pointed by the handle must be FROZEN.
	*/
	YSRESULT DeleteFrozen(YsEditArrayObjectHandle<T,bitShift> hd);

	/*! Find an object from a search key.  It can find an object that is ALIVE.
	*/
	YsEditArrayObjectHandle<T,bitShift> FindObject(unsigned int key) const;

	/*! Find a frozen object from a search key.
	*/
	YsEditArrayObjectHandle<T,bitShift> FindFrozenObject(unsigned int key) const;

	/*! Find next object of obHd that is ALIVE.  The object pointed by obHd does not have to be ALIVE.
	    If obHd==NULL, this function finds the next first object ALIVE and returns the handle to the next object.
	*/
	YsEditArrayObjectHandle <T,bitShift> FindNext(const YsEditArrayObjectHandle <T,bitShift> obHd) const;

	/*! Find previous object obHd that is ALIVE.  The object pointed by obHd does not have to be ALIVE.
	    If obHd==NULL, this function finds the last object ALIVE and returns the handle to the previous object.
	*/
	YsEditArrayObjectHandle <T,bitShift> FindPrev(const YsEditArrayObjectHandle <T,bitShift> obHd) const;

	/*! Move obHd to the next object that is ALIVE.  The object pointed by obHd does not have to be ALIVE.
	    If obHd==NULL, this function finds the next first object ALIVE.
	    If the next of obHd exists, this function returns YSOK.
	    If the next of obHd does not exist, this function returns YSERR, and obHd will be NULL.
	*/
	YSRESULT MoveToNext(YsEditArrayObjectHandle <T,bitShift> &obHd) const;

	/*! Move obHd to the previous object that is ALIVE.  The object pointed by obHd does not have to be ALIVE.
	    If obHd==NULL, this function finds the last object ALIVE.
	    If the previous of obHd exists, this function returns YSOK, and obHd will be updated to the handle to the previous object.
	    If the previous of obHd does not exist, this function returns YSERR, and obHd will be NULL.
	*/
	YSRESULT MoveToPrev(YsEditArrayObjectHandle <T,bitShift> &obHd) const;
};

template <class T,const int bitShift>
YsEditArray <T,bitShift>::YsEditArray()
{
	localSearchKeySeed=0;
	searchKeySeed=&localSearchKeySeed;

	searchEnabled=YSFALSE;
	nFrozen=0;
}

template <class T,const int bitShift>
void YsEditArray <T,bitShift>::CleanUp(void)
{
	objArray.Clear();
	deletedIndex.Clear();
	nFrozen=0;
	searchKeyToIndex.PrepareTable();;
}

template <class T,const int bitShift>
void YsEditArray <T,bitShift>::UseSharedSearchKeySeed(unsigned int &seedPtr)
{
	searchKeySeed=&seedPtr;
}

template <class T,const int bitShift>
void YsEditArray <T,bitShift>::EnableSearch(void)
{
	if(YSTRUE!=searchEnabled)
	{
		searchEnabled=YSTRUE;
		searchKeyToIndex.PrepareTable(1+objArray.GetN()/4);

		for(YsEditArrayObjectHandle<T,bitShift> ptr=NULL; YSOK==MoveToNext(ptr); )
		{
			YSSIZE_T index=ptr.index;
			searchKeyToIndex.AddElement(objArray[index].GetSearchKey(),index);
		}
	}
}

template <class T,const int bitShift>
void YsEditArray <T,bitShift>::DisableSearch(void)
{
	searchEnabled=YSFALSE;
}

template <class T,const int bitShift>
YSSIZE_T YsEditArray <T,bitShift>::GetN(void) const
{
	return objArray.GetN()-deletedIndex.GetN()-nFrozen;
}

template <class T,const int bitShift>
YSBOOL YsEditArray <T,bitShift>::IsFrozen(YsEditArrayObjectHandle<T,bitShift> obHd) const
{
	if(NULL!=obHd &&
	   YSTRUE==objArray.IsInRange(obHd.index) &&
	   FROZEN==objArray[obHd.index].GetState())
	{
		return YSTRUE;
	}
	return YSFALSE;
}

template <class T,const int bitShift>
YsEditArrayObjectHandle<T,bitShift> YsEditArray <T,bitShift>::Create(void)
{
	YSSIZE_T newIdx;
	if(0<deletedIndex.GetN())
	{
		newIdx=deletedIndex.GetEnd();
		deletedIndex.DeleteLast();
	}
	else
	{
		newIdx=objArray.GetN();
		objArray.Increment();
	}

	YsEditArrayObjectHandle<T,bitShift> obHd;
	obHd.index=newIdx;

	objArray[newIdx].SetState(ALIVE);
	objArray[newIdx].SetSearchKey(*searchKeySeed);

	if(YSTRUE==searchEnabled)
	{
		searchKeyToIndex.AddElement(*searchKeySeed,newIdx);
	}

	++(*searchKeySeed);

	return obHd;
}

template <class T,const int bitShift>
T *YsEditArray <T,bitShift>::operator[] (YsEditArrayObjectHandle<T,bitShift> hd)
{
	if(YSTRUE==objArray.IsInRange(hd.index) &&
	   ALIVE==objArray[hd.index].GetState())
	{
		return &objArray[hd.index].dat;
	}
	return NULL;
}

template <class T,const int bitShift>
const T *YsEditArray <T,bitShift>::operator[] (YsEditArrayObjectHandle<T,bitShift> hd) const
{
	if(YSTRUE==objArray.IsInRange(hd.index) &&
	   ALIVE==objArray[hd.index].GetState())
	{
		return &objArray[hd.index].dat;
	}
	return NULL;
}

template <class T,const int bitShift>
const T *YsEditArray <T,bitShift>::GetFrozenObject(YsEditArrayObjectHandle<T,bitShift> hd) const
{
	if(YSTRUE==objArray.IsInRange(hd.index) &&
	   FROZEN==objArray[hd.index].GetState())
   {
		return &objArray[hd.index].dat;
	}
	return NULL;
}

template <class T,const int bitShift>
YsEditArrayObjectHandle<T,bitShift> YsEditArray <T,bitShift>::Append(const T &incoming)
{
	YsEditArrayObjectHandle<T,bitShift> newObjHd=Create();
	(*(*this)[newObjHd])=incoming;
	return newObjHd;
}

template <class T,const int bitShift>
YSRESULT YsEditArray <T,bitShift>::Delete(YsEditArrayObjectHandle<T,bitShift> obHd)
{
	if(NULL!=obHd && 
	   YSTRUE==objArray.IsInRange(obHd.index) &&
	   ALIVE==objArray[obHd.index].GetState())
	{
		searchKeyToIndex.DeleteElement(objArray[obHd.index].GetSearchKey(),obHd.index);
		objArray[obHd.index].SetState(DELETED);
		deletedIndex.Append(obHd.index);
		return YSOK;
	}
	return YSERR;
}

template <class T,const int bitShift>
YSRESULT YsEditArray <T,bitShift>::Freeze(YsEditArrayObjectHandle<T,bitShift> obHd)
{
	if(NULL!=obHd &&
	   YSTRUE==objArray.IsInRange(obHd.index) &&
	   ALIVE==objArray[obHd.index].GetState())
	{
		objArray[obHd.index].SetState(FROZEN);
		++nFrozen;
		return YSOK;
	}
	return YSERR;
}

template <class T,const int bitShift>
YSRESULT YsEditArray <T,bitShift>::Melt(YsEditArrayObjectHandle<T,bitShift> obHd)
{
	if(NULL!=obHd &&
	   YSTRUE==objArray.IsInRange(obHd.index) &&
	   FROZEN==objArray[obHd.index].GetState())
	{
		objArray[obHd.index].SetState(ALIVE);
		--nFrozen;
		return YSOK;
	}
	return YSERR;
}

template <class T,const int bitShift>
YSRESULT YsEditArray <T,bitShift>::DeleteFrozen(YsEditArrayObjectHandle<T,bitShift> obHd)
{
	if(NULL!=obHd &&
	   YSTRUE==objArray.IsInRange(obHd.index) &&
	   FROZEN==objArray[obHd.index].GetState())
	{
		Melt(obHd);
		Delete(obHd);
		return YSOK;
	}
	return YSERR;
}

template <class T,const int bitShift>
YsEditArrayObjectHandle<T,bitShift> YsEditArray <T,bitShift>::FindObject(unsigned int key) const
{
	YSSIZE_T index;
	if(YSOK==searchKeyToIndex.FindElement(index,key) &&
	   YSTRUE==objArray.IsInRange(index) &&
	   ALIVE==objArray[index].GetState())
	{
		YsEditArrayObjectHandle<T,bitShift> obHd;
		obHd.index=index;
		return obHd;
	}
	return NULL;
}

template <class T,const int bitShift>
YsEditArrayObjectHandle<T,bitShift> YsEditArray <T,bitShift>::FindFrozenObject(unsigned int key) const
{
	YSSIZE_T index;
	if(YSOK==searchKeyToIndex.FindElement(index,key) &&
	   YSTRUE==objArray.IsInRange(index) &&
	   FROZEN==objArray[index].GetState())
	{
		YsEditArrayObjectHandle<T,bitShift> obHd;
		obHd.index=index;
		return obHd;
	}
	return NULL;
}

template <class T,const int bitShift>
YsEditArrayObjectHandle <T,bitShift> YsEditArray <T,bitShift>::FindNext(const YsEditArrayObjectHandle <T,bitShift> obHd) const
{
	YsEditArrayObjectHandle <T,bitShift> nextObHd=obHd;
	if(YSOK==MoveToNext(nextObHd))
	{
		return nextObHd;
	}
	return NULL;
}

template <class T,const int bitShift>
YsEditArrayObjectHandle <T,bitShift> YsEditArray <T,bitShift>::FindPrev(const YsEditArrayObjectHandle <T,bitShift> obHd) const
{
	YsEditArrayObjectHandle <T,bitShift> prevObHd=obHd;
	if(YSOK==MoveToPrev(prevObHd))
	{
		return prevObHd;
	}
	return NULL;
}

template <class T,const int bitShift>
YSRESULT YsEditArray <T,bitShift>::MoveToNext(YsEditArrayObjectHandle <T,bitShift> &obHd) const
{
	YSSIZE_T i0;
	if(NULL==obHd || YSTRUE!=objArray.IsInRange(obHd.index))
	{
		i0=0;
	}
	else
	{
		i0=obHd.index+1;
	}
	for(YSSIZE_T i=i0; i<objArray.GetN(); ++i)
	{
		if(ALIVE==objArray[i].GetState())
		{
			obHd.index=i;
			return YSOK;
		}
	}
	obHd=NULL;
	return YSERR;
}

template <class T,const int bitShift>
YSRESULT YsEditArray <T,bitShift>::MoveToPrev(YsEditArrayObjectHandle <T,bitShift> &obHd) const
{
	YSSIZE_T i0;
	if(NULL==obHd || YSTRUE!=objArray.IsInRange(obHd.index))
	{
		i0=objArray.GetN()-1;
	}
	else
	{
		i0=obHd.index-1;
	}
	for(int i=i0; 0<=i; --i)
	{
		if(ALIVE==objArray[i].GetState())
		{
			obHd.index=i;
			return YSOK;
		}
	}
	obHd=NULL;
	return YSERR;
}

/* } */
#endif
