#include "yskeyword.h"

YsKeyWordList::YsKeyWordList() : keyWordList(keyWordAlloc)
{
}

void YsKeyWordList::Initialize(void)
{
	keyWordList.CleanUp();
	keyWordDict.Set(0,NULL);
}

int YsKeyWordList::GetN(void) const
{
	return keyWordList.GetN();
}

YSRESULT YsKeyWordList::MakeList(const char * const str[],const YSSIZE_T attrib[])
{
	int i,n;
	YsListItem <YsKeyWord> *kw;
	YsArray <const char *> ptr;

	Initialize();

	for(i=0; str[i]!=NULL; i++)
	{
		kw=keyWordList.Create();
		kw->dat.str.Set(str[i]);
		kw->dat.id=i;
		kw->dat.attrib=(NULL!=attrib ? attrib[i] : 0);

		ptr.Append(str[i]);
		keyWordDict.Append(&kw->dat);
	}
	n=i;

	YsQuickSortString <YsKeyWord *> (n,ptr,keyWordDict);

	return YSOK;
}

YSRESULT YsKeyWordList::AddKeyWord(int nStr,const char * const str[],const YSSIZE_T attrib[])
{
	int i,j,k,id;
	YsListItem <YsKeyWord> *kw;
	YsArray <YsKeyWord *> prvKwd,newKwd;
	YsArray <const char *> newKwdPtr;

	id=keyWordList.GetN();
	newKwd.Set(nStr,NULL);
	newKwdPtr.Set(nStr,NULL);
	for(i=0; i<nStr; i++)
	{
		kw=keyWordList.Create();
		kw->dat.str.Set(str[i]);
		kw->dat.id=id;
		kw->dat.attrib=(NULL!=attrib ? attrib[i] : 0);

		newKwd[i]=&kw->dat;
		newKwdPtr[i]=kw->dat.str;

		id++;
	}

	YsQuickSortString <YsKeyWord *> (newKwdPtr.GetN(),newKwdPtr,newKwd);

	prvKwd=keyWordDict;

	keyWordDict.Set(prvKwd.GetN()+newKwd.GetN(),NULL);

	i=0;
	j=0;
	k=0;
	while(i<prvKwd.GetN() || j<newKwd.GetN())
	{
		if(newKwd.GetN()<=j)
		{
			keyWordDict[k++]=prvKwd[i++];
		}
		else if(prvKwd.GetN()<=i)
		{
			keyWordDict[k++]=newKwd[j++];
		}
		else if(strcmp(prvKwd[i]->str,newKwd[j]->str)<0)
		{
			keyWordDict[k++]=prvKwd[i++];
		}
		else
		{
			keyWordDict[k++]=newKwd[j++];
		}
	}

	return YSOK;
}

YSRESULT YsKeyWordList::AddKeyWord(int nStr,const YsString str[],const YSSIZE_T attrib[])
{
	YsArray <const char *> ptr;
	ptr.Set(nStr,NULL);
	for(int i=0; i<nStr; i++)
	{
		ptr[i]=str[i];
	}
	return AddKeyWord(nStr,ptr.GetArray(),attrib);
}

int YsKeyWordList::GetId(const char str[]) const
{
	return FindIndex(str);
}

int YsKeyWordList::FindIndex(const char str[]) const
{
	if(keyWordDict.GetN()==0)
	{
		return -1;
	}

	YSSIZE_T idLow,idHigh;
	idLow=0;
	idHigh=keyWordDict.GetN()-1;

	while(idHigh-idLow>1)
	{
		YSSIZE_T idMid,dif;
		idMid=idLow+idHigh;
		idMid>>=1;
		dif=strcmp(str,keyWordDict[idMid]->str);
		if(dif==0)
		{
			return keyWordDict[idMid]->id;
		}
		else if(dif<0)
		{
			idHigh=idMid;
		}
		else
		{
			idLow=idMid;
		}
	}

	if(strcmp(str,keyWordDict[idLow]->str)==0)
	{
		return keyWordDict[idLow]->id;
	}
	else if(strcmp(str,keyWordDict[idHigh]->str)==0)
	{
		return keyWordDict[idHigh]->id;
	}

	return -1;
}

const char *YsKeyWordList::GetString(int id) const
{
	if(keyWordDict.IsInRange(id)==YSTRUE)
	{
		return keyWordDict[id]->str;
	}
	return NULL;
}

YSSIZE_T YsKeyWordList::GetAttribute(int idx) const
{
	if(keyWordDict.IsInRange(idx)==YSTRUE)
	{
		return keyWordDict[idx]->attrib;
	}
	return -1;
}

