#if (defined(_WIN32) || defined(_WIN64)) && !defined(_CRT_SECURE_NO_WARNINGS)
	// This disables annoying warnings VC++ gives for use of C standard library.  Shut the mouth up.
	#define _CRT_SECURE_NO_WARNINGS
#endif

#include <stdio.h>
#include <stdlib.h>
#include "ystextresource.h"

class YsTextResourceBinaryTree : public YsStringBinaryTree <YsString,YsWString>
{
};

YsTextResource::YsTextResource()
{
	tree=new YsTextResourceBinaryTree;
	Initialize();
}

YsTextResource::~YsTextResource()
{
	delete tree;
}

void YsTextResource::Initialize(void)
{
	tree->CleanUp();
}


YSRESULT YsTextResource::LoadFile(const char fn[])
{
	FILE *fp=fopen(fn,"r");
	YSRESULT res=LoadFile(fp);
	if(NULL!=fp)
	{
		fclose(fp);
	}
	return res;
}

YSRESULT YsTextResource::LoadFile(FILE *fp)
{
	if(NULL!=fp)
	{
		Initialize();

		YSRESULT res=YSOK;
		YsString buf;
		while(NULL!=buf.Fgets(fp))
		{
			while(0<buf.Strlen() && buf.LastChar()=='\n' || buf.LastChar()=='\r')
			{
				buf.BackSpace();
			}

			if(YSOK!=SendCommand(buf))
			{
				res=YSERR;
			}
		}
		return res;
	}
	return YSERR;
}

YSRESULT YsTextResource::SendCommand(const YsString &cmd)
{
	if(0==cmd.Strlen())
	{
		return YSOK;
	}

	switch(cmd[0])
	{
	case '*':
		return YSOK;
	case '$':
		return SendDollarCommand(cmd);
	case '+':
		return SendDictCommand(cmd);
	}
	return YSOK;
}

YSRESULT YsTextResource::SendDollarCommand(const YsString &cmd)
{
	YsArray <YsString,16> args;
	cmd.Arguments <16> (args);
	if(0<args.GetN())
	{
		args[0].Capitalize();
		if(0==strcmp(args[0],"$FGCOL"))
		{
			if(2==args.GetN())
			{
				fgCol.SetIntRGB(atoi(args[1]),atoi(args[1]),atoi(args[1]));
			}
			else if(4<=args.GetN())
			{
				fgCol.SetIntRGB(atoi(args[1]),atoi(args[2]),atoi(args[3]));
			}
		}
		else if(0==strcmp(args[0],"$BGCOL"))
		{
			if(2==args.GetN())
			{
				bgCol.SetIntRGB(atoi(args[1]),atoi(args[1]),atoi(args[1]));
			}
			else if(4<=args.GetN())
			{
				bgCol.SetIntRGB(atoi(args[1]),atoi(args[2]),atoi(args[3]));
			}
		}
		else if(0==strcmp(args[0],"$LABEL"))
		{
			if(2<=args.GetN())
			{
				label=args[1];
			}
		}
	}
	return YSOK;
}

YSRESULT YsTextResource::SendDictCommand(const YsString &cmd)
{
	YsString key,msg;
	int i;
	for(i=1; i<cmd.Strlen(); i++)
	{
		if('#'==cmd[i])
		{
			break;
		}
		key.Append(cmd[i]);
	}

	if('#'!=cmd[i])
	{
		return YSERR;
	}

	for(i=i+1; i<cmd.Strlen(); i++)
	{
		if('\\'==cmd[i] && 'n'==cmd[i+1])
		{
			msg.Append('\n');
			i++;
		}
		else if('\\'==cmd[i] && 't'==cmd[i+1])
		{
			msg.Append('\t');
			i++;
		}
		else
		{
			msg.Append(cmd[i]);
		}
	}

	key.DeleteTailSpace();

	YsWString wMsg;
	wMsg.SetUTF8String(msg);

	tree->AddNode(key,wMsg);

	return YSOK;
}

const wchar_t *YsTextResource::FindWString(const char key[])
{
	YsString ysKey(key);
	const YsWString *msg=tree->FindData(ysKey);
	if(NULL!=msg)
	{
		return msg->Txt();
	}
	return NULL;
}
