#include "controldialog.h"


StandbyDialog *StandbyDialog::Create(void)
{
	return new StandbyDialog();
}

void StandbyDialog::Delete(StandbyDialog *ptr)
{
	delete ptr;
}

void StandbyDialog::Make(FsGui3DMainCanvas *mainCanvas)
{
	this->mainCanvas=mainCanvas;
	openButton=AddTextButton(0,FSKEY_NULL,FSGUI_PUSHBUTTON,"Open Control Dialog",YSTRUE);
	Fit();
	SetBackgroundAlpha(0.0);
	SetArrangeType(FSDIALOG_ARRANGE_BOTTOM_LEFT);
}

void StandbyDialog::OnButtonClick(FsGuiButton *btn)
{
	if(btn==openButton)
	{
		mainCanvas->RemoveDialog(this);
		mainCanvas->AddDialog(mainCanvas->controlDlg);
		mainCanvas->ArrangeDialog();
	}
}



ControlDialog *ControlDialog::Create(void)
{
	return new ControlDialog;
}

void ControlDialog::Delete(ControlDialog *ptr)
{
	delete ptr;
}

const double ControlDialog::GetThrottle(void) const
{
	return throttleSlider->GetScaledValue();
}

const YSBOOL ControlDialog::GetLight(void) const
{
	return lightBtn->GetCheck();
}

void ControlDialog::Make(FsGui3DMainCanvas *mainCanvas)
{
	this->mainCanvas=mainCanvas;

	bgColorPalette=AddColorPalette(0,FSKEY_NULL,"Background",255,255,255,5,YSTRUE,YSTRUE);

	afterBurnerBtn=AddTextButton(0,FSKEY_NULL,FSGUI_CHECKBOX,"Afterburner",YSTRUE);
	lightBtn=AddTextButton(0,FSKEY_NULL,FSGUI_CHECKBOX,"Light",YSTRUE);

	YsArray <YsArray <FsGuiDialogItem *> > dlgItemTable;

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Elevator",YSTRUE));
	elevatorSlider=AddHorizontalSlider(0,FSKEY_NULL,20,-1.0,1.0,YSFALSE);
	elevatorSlider->SetPosition(0.5);
	dlgItemTable.GetEnd().Append(elevatorSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Aileron",YSTRUE));
	aileronSlider=AddHorizontalSlider(0,FSKEY_NULL,20,-1.0,1.0,YSFALSE);
	aileronSlider->SetPosition(0.5);
	dlgItemTable.GetEnd().Append(aileronSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Rudder",YSTRUE));
	rudderSlider=AddHorizontalSlider(0,FSKEY_NULL,20,-1.0,1.0,YSFALSE);
	rudderSlider->SetPosition(0.5);
	dlgItemTable.GetEnd().Append(rudderSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Throttle",YSTRUE));
	throttleSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(throttleSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Flaps",YSTRUE));
	flapSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(flapSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Landing Gear",YSTRUE));
	gearSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(gearSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Spoiler/AirBrake",YSTRUE));
	spoilerSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(spoilerSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Brake/Hook",YSTRUE));
	brakeSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(brakeSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Thrust Reverser",YSTRUE));
	reverserSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(reverserSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Variable Geometry Wing",YSTRUE));
	vgwSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(vgwSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Nozzle/Nose",YSTRUE));
	nozzleSlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(nozzleSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Weapon Bay",YSTRUE));
	bombBaySlider=AddHorizontalSlider(0,FSKEY_NULL,20,0.0,1.0,YSFALSE);
	dlgItemTable.GetEnd().Append(bombBaySlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Turret (H)",YSTRUE));
	turretHSlider=AddHorizontalSlider(0,FSKEY_NULL,20,-YsPi,YsPi,YSFALSE);
	turretHSlider->SetPosition(0.5);
	dlgItemTable.GetEnd().Append(turretHSlider);

	dlgItemTable.Increment();
	dlgItemTable.GetEnd().Append(AddStaticText(0,FSKEY_NULL,"Turret (P)",YSTRUE));
	turretPSlider=AddHorizontalSlider(0,FSKEY_NULL,20,-YsPi/2.0,YsPi/2.0,YSFALSE);
	turretPSlider->SetPosition(0.5);
	dlgItemTable.GetEnd().Append(turretPSlider);

	AlignLeftMiddle(dlgItemTable);

	closeButton=AddTextButton(0,FSKEY_NULL,FSGUI_PUSHBUTTON,"Close",YSTRUE);
	Fit();
	SetBackgroundAlpha(0.0);
	SetArrangeType(FSDIALOG_ARRANGE_BOTTOM_LEFT);
}

void ControlDialog::OnButtonClick(FsGuiButton *btn)
{
	if(btn==closeButton)
	{
		mainCanvas->RemoveDialog(this);
		mainCanvas->AddDialog(mainCanvas->standbyDlg);
		mainCanvas->ArrangeDialog();
	}
	else
	{
		if(FsGui3DMainCanvas::RENDER_DNM==mainCanvas->whatToDraw)
		{
			if(btn==afterBurnerBtn)
			{
				if(YSTRUE==btn->GetCheck())
				{
					mainCanvas->dnm.SetClassState(DNMCLASS_AIR_AFTERBURNER,1);
				}
				else
				{
					mainCanvas->dnm.SetClassState(DNMCLASS_AIR_AFTERBURNER,0);
				}
			}
			else if(btn==lightBtn)
			{
				if(YSTRUE==btn->GetCheck())
				{
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_NAVLIGHT,YSTRUE);
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_BEACON,YSTRUE);
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_STROBE,YSTRUE);
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_LANDINGLIGHT,YSTRUE);
					if(YsTolerance<gearSlider->GetScaledValue())
					{
						mainCanvas->dnm.SetShow(DNMCLASS_AIR_LIGHTONLANDINGGEAR,YSTRUE);
					}
				}
				else
				{
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_NAVLIGHT,YSFALSE);
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_BEACON,YSFALSE);
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_STROBE,YSFALSE);
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_LANDINGLIGHT,YSFALSE);
					mainCanvas->dnm.SetShow(DNMCLASS_AIR_LIGHTONLANDINGGEAR,YSFALSE);
				}
			}
		}
	}
}

void ControlDialog::OnColorPaletteChange(FsGuiColorPalette *plt)
{
	plt->GetColor(mainCanvas->bgCol);
}

void ControlDialog::OnSliderPositionChange(FsGuiSlider *slider,const double & /*prevPos*/,const double & /*prevValue*/)
{
	if(FsGui3DMainCanvas::RENDER_DNM==mainCanvas->whatToDraw)
	{
		const double t=slider->GetScaledValue();

		if(slider==gearSlider)
		{
			if(YsTolerance<t)
			{
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_TIRE,YSTRUE);
			}
			else
			{
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_TIRE,YSFALSE);
			}

			if(t<0.2)
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_LANDINGGEAR,0,1,0.0);
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_GEARDOOR,0,1,t/0.2);
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_GEARDOOR_FAST,0,1,t/0.2);
			}
			else if(t<0.8)
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_LANDINGGEAR,0,1,(t-0.2)/0.6);
				mainCanvas->dnm.SetClassState(DNMCLASS_AIR_GEARDOOR,1);
				mainCanvas->dnm.SetClassState(DNMCLASS_AIR_GEARDOOR_FAST,1);
			}
			else
			{
				mainCanvas->dnm.SetClassState(DNMCLASS_AIR_LANDINGGEAR,1);
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_GEARDOOR,0,1,(1.0-t)/0.2);
				mainCanvas->dnm.SetClassState(DNMCLASS_AIR_GEARDOOR_FAST,1);
			}

			if(t<0.6)
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_GEARROOMWALL,0,1,t/0.6);
			}
			else
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_GEARROOMWALL,0,1,1.0);
			}

			if(/*YSTRUE==staLandingLight &&*/ t>0.2)
			{
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_LIGHTONLANDINGGEAR,YSTRUE);
			}
			else
			{
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_LIGHTONLANDINGGEAR,YSFALSE);
			}

			mainCanvas->SetNeedRedraw(YSTRUE);
		}
		else if(slider==nozzleSlider)
		{
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_NOZZLE,0,1,t);
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_CONCORDENOSE,0,1,t);
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_CONCORDEVISOR,0,1,YsBound(t*2.5,0.0,1.0));

			mainCanvas->SetNeedRedraw(YSTRUE);
		}
		else if(slider==throttleSlider)
		{
			if(t>0.6)
			{
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_BLADE_FAST,YSTRUE);  // Fast
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_BLADE_SLOW,YSFALSE); // Slow
			}
			else if(t>0.3)
			{
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_BLADE_FAST,YSTRUE);  // Fast
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_BLADE_SLOW,YSTRUE);  // Slow
			}
			else
			{
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_BLADE_FAST,YSFALSE); // Fast
				mainCanvas->dnm.SetShow(DNMCLASS_AIR_BLADE_SLOW,YSTRUE);  // Slow
			}
		}
		else if(slider==flapSlider)
		{
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_FLAP,0,1,t);
		}
		else if(slider==elevatorSlider)
		{
			if(t>=0.0)
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_ELEVATOR,0,1,t);
			}
			else
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_ELEVATOR,0,2,-t);
			}
		}
		else if(slider==aileronSlider)
		{
			if(t>=0.0)
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_AILERON,0,1,t);
			}
			else
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_AILERON,0,2,-t);
			}
		}
		else if(slider==rudderSlider)
		{
			if(t>=0.0)
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_RUDDER,0,1,t);
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_STEERING,0,1,t);
			}
			else
			{
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_RUDDER,0,2,-t);
				mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_STEERING,0,2,-t);
			}
		}
		else if(slider==spoilerSlider)
		{
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_AIRBRAKE,0,1,t);
		}
		else if(slider==brakeSlider)
		{
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_BRAKE_HOOK,0,1,t);
		}
		else if(slider==reverserSlider)
		{
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_THRUSTREVERSER,0,1,t);
		}
		else if(slider==vgwSlider)
		{
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_VGW,0,1,t);
		}
		else if(slider==bombBaySlider)
		{
			mainCanvas->dnm.InterpolateClassState(DNMCLASS_AIR_BOMBBAY,0,1,t);
		}
		else if(slider==turretHSlider)
		{
			mainCanvas->dnm.SetClassHeading(DNMCLASS_AIR_TURRET,t);
		}
		else if(slider==turretPSlider)
		{
			mainCanvas->dnm.SetClassPitch(DNMCLASS_AIR_TURRET,t);
		}
	}
}

