#include "fsgui3dapp.h"
#include "controldialog.h"

#include <ysgl.h>

static YSBOOL appMustTerminate=YSFALSE;
static int appSleepInterval=10;
static FsGui3DMainCanvas *appMainCanvas;


////////////////////////////////////////////////////////////

FsGui3DMainMenu::FsGui3DMainMenu(class FsGui3DMainCanvas *canvas)
{
	this->canvas=canvas;
}

void FsGui3DMainMenu::OnSelectMenuItem(FsGuiPopUpMenuItem *item)
{
	canvas->OnSelectMenuItem(item);
}

////////////////////////////////////////////////////////////

FsGui3DMainCanvas::FsGui3DMainCanvas() : threeDInterface(this),viewControlDlg(NULL)
{
	MakeMainMenu();
	AddViewControlDialog();
	drawEnv.AttachCanvas(this);

	standbyDlg=StandbyDialog::Create();
	standbyDlg->Make(this);
	
	controlDlg=ControlDialog::Create();
	controlDlg->Make(this);

	bgCol=YsWhite();

	AddDialog(controlDlg);

	whatToDraw=RENDER_NONE;
	t=0.0;
	night=YSFALSE;
}

FsGui3DMainCanvas::~FsGui3DMainCanvas()
{
	ControlDialog::Delete(controlDlg);
	StandbyDialog::Delete(standbyDlg);
	if(NULL!=viewControlDlg)
	{
		FsGui3DViewControlDialog::Delete(viewControlDlg);
	}
	DeleteMainMenu();
}

void FsGui3DMainCanvas::AddViewControlDialog(void)
{
	viewControlDlg=FsGui3DViewControlDialog::Create();
	viewControlDlg->Make(this,&drawEnv);
	this->AddDialog(viewControlDlg);
}

void FsGui3DMainCanvas::MakeMainMenu(void)
{
	mainMenu=new FsGui3DMainMenu(this);
	mainMenu->Initialize();
	mainMenu->SetIsPullDownMenu(YSTRUE);

	FsGuiPopUpMenuItem *fileMenu=mainMenu->AddTextItem(0,FSKEY_F,L"File(F)");
	FsGuiPopUpMenu *fileSubMenu=fileMenu->GetSubMenu();
	fileSubMenu->AddTextItem(MENU_FILE_OPEN,FSKEY_O,L"Open(O)");
	fileSubMenu->AddTextItem(MENU_FILE_EXIT,FSKEY_X,L"Exit(X)");

	SetMainMenu(mainMenu);
}

void FsGui3DMainCanvas::DeleteMainMenu(void)
{
	delete mainMenu;
}

void FsGui3DMainCanvas::OnInterval(void)
{
	const double dt=(double)FsPassedTime()/1000.0;

	{
		int key;
		while(FSKEY_NULL!=(key=FsInkey()))
		{
			if(YSOK!=drawEnv.KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT)))
			{
				if(YSOK!=threeDInterface.KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT)))
				{
					this->KeyIn(key,(YSBOOL)FsGetKeyState(FSKEY_SHIFT),(YSBOOL)FsGetKeyState(FSKEY_CTRL),(YSBOOL)FsGetKeyState(FSKEY_ALT));
				}
			}
		}
	}

	{
		int charCode;
		while(0!=(charCode=FsInkeyChar()))
		{
			this->CharIn(charCode);
		}
	}

	{
		int lb,mb,rb,mx,my;
		while(FSMOUSEEVENT_NONE!=FsGetMouseEvent(lb,mb,rb,mx,my))
		{
			if(YSOK!=drawEnv.SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
			{
				if(YSOK!=this->SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
				{
					if(YSOK!=threeDInterface.SetMouseState((YSBOOL)lb,(YSBOOL)mb,(YSBOOL)rb,mx,my))
					{
					}
				}
			}
		}
	}

	if((YsTolerance<controlDlg->GetThrottle() || YSTRUE==controlDlg->GetLight()) && RENDER_DNM==whatToDraw)
	{
		t+=dt;
		if(YsTolerance<controlDlg->GetThrottle())
		{
			const double prop=t*controlDlg->GetThrottle()*100.0;
			dnm.SetClassBank(DNMCLASS_AIR_ROTOR,prop);
			const double tire=t*controlDlg->GetThrottle()*10.0;
			dnm.SetClassPitch(DNMCLASS_AIR_TIRE,-tire);
		}
		if(YSTRUE==controlDlg->GetLight())
		{
			const YSBOOL beacon=(1.0<fmod(t,2.0) ? YSTRUE :YSFALSE);
			dnm.SetShow(DNMCLASS_AIR_BEACON,beacon);

			const YSBOOL strobe=(0.1>fmod(t,1.0) ? YSTRUE : YSFALSE);
			dnm.SetShow(DNMCLASS_AIR_STROBE,strobe);
		}
		SetNeedRedraw(YSTRUE);
	}



	int winWid,winHei;
	FsGetWindowSize(winWid,winHei);
	this->SetWindowSize(winWid,winHei,/*autoArrangeDialog=*/YSTRUE);

	drawEnv.SetWindowSize(winWid,winHei);
	drawEnv.SetViewportByTwoCorner(0,0,winWid,winHei);


	if(YSTRUE==NeedRedraw() ||
	   0!=FsCheckWindowExposure())
	{
		Draw();
	}
}

void FsGui3DMainCanvas::OnSelectMenuItem(FsGuiPopUpMenuItem *menuItem)
{
	if(NULL!=menuItem)
	{
		switch(menuItem->GetId())
		{
		case MENU_FILE_EXIT:
			appMustTerminate=YSTRUE;
			break;
		case MENU_FILE_OPEN:
			FileOpen();
			break;
		}
	}
}

void FsGui3DMainCanvas::Draw(void)
{
	int viewport[4];
	drawEnv.GetOpenGlCompatibleViewport(viewport);
	drawEnv.SetVerticalOrientation(YSTRUE);
	drawEnv.UpdateNearFar();


	glUseProgram(0);


	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);


	const YsMatrix4x4 &projMat=drawEnv.GetProjectionMatrix();
	const YsMatrix4x4 &viewMat=drawEnv.GetViewMatrix();
	const YsMatrix4x4 projViewMat=(projMat*viewMat);

	threeDInterface.SetViewport(viewport);
	threeDInterface.SetProjViewModelMatrix(projViewMat);


	glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);

	GLfloat projection[16];
	projMat.GetOpenGlCompatibleMatrix(projection);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMultMatrixf(projection);

	GLfloat modelview[16];
	viewMat.GetOpenGlCompatibleMatrix(modelview);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMultMatrixf(modelview);

	glClearColor(bgCol.Rf(),bgCol.Gf(),bgCol.Bf(),0.0);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	threeDInterface.Draw();



	if(RENDER_NONE!=whatToDraw)
	{
		YsMatrix4x4 correctedViewMat;  // Corrected for left-hand coordinate system.
		correctedViewMat.Scale(1.0,1.0,-1.0);
		correctedViewMat*=viewMat;
		correctedViewMat.Scale(1.0,1.0,-1.0);

		glEnable(GL_CULL_FACE);

		GLfloat modelview[16];
		correctedViewMat.GetOpenGlCompatibleMatrix(modelview);

		YsGLSLUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		YsGLSLSet3DRendererProjectionfv(YsGLSLSharedVariColorShaded3DRenderer(),projection);
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColorShaded3DRenderer(),modelview);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColorShaded3DRenderer());
		
		YsGLSLUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		YsGLSLSet3DRendererProjectionfv(YsGLSLSharedVariColor3DRenderer(),projection);
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedVariColor3DRenderer(),modelview);
		YsGLSLEndUse3DRenderer(YsGLSLSharedVariColor3DRenderer());
		
		YsGLSLUse3DRenderer(YsGLSLSharedFlat3DRenderer());
		YsGLSLSet3DRendererProjectionfv(YsGLSLSharedFlat3DRenderer(),projection);
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedFlat3DRenderer(),modelview);
		YsGLSLEndUse3DRenderer(YsGLSLSharedFlat3DRenderer());


		YsGLSLUse3DRenderer(YsGLSLSharedFlash3DRenderer());
		YsGLSLSet3DRendererProjectionfv(YsGLSLSharedFlash3DRenderer(),projection);
		YsGLSLSet3DRendererModelViewfv(YsGLSLSharedFlash3DRenderer(),modelview);
		if(YSTRUE!=night)
		{
			YsGLSLSet3DRendererUniformFlashSize(YsGLSLSharedFlash3DRenderer(),0.1f);
			YsGLSLSet3DRendererFlashRadius(YsGLSLSharedFlash3DRenderer(),0.6f,1.0f);
		}
		else
		{
			YsGLSLSet3DRendererUniformFlashSize(YsGLSLSharedFlash3DRenderer(),1.0f);
			YsGLSLSet3DRendererFlashRadius(YsGLSLSharedFlash3DRenderer(),0.2f,1.0f);
		}
		YsGLSLEndUse3DRenderer(YsGLSLSharedFlash3DRenderer());


		double nearZ,farZ;
		drawEnv.GetNearFar(nearZ,farZ);
		const double tanFov=tan(drawEnv.GetFOVY());

		const YsVec3 pos=YsOrigin();
		const YsAtt3 att=YsZeroAtt();

		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_DEPTH_TEST);

		switch(whatToDraw)
		{
		case RENDER_DNM:
		case RENDER_SRF:
			dnm.DrawVisual(correctedViewMat,pos,att,nearZ,farZ,tanFov,YSDNM_DRAWOPAQUE);
			dnm.DrawVisual(correctedViewMat,pos,att,nearZ,farZ,tanFov,YSDNM_DRAWTRANSPARENT);
			break;
		}
	}


	// 2D Drawing
	glDisable(GL_CULL_FACE);

	YsGLSLUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());
	YsGLSLUseWindowCoordinateInPlain2DDrawing(YsGLSLSharedPlain2DRenderer(),YSTRUE);
	YsGLSLEndUsePlain2DRenderer(YsGLSLSharedPlain2DRenderer());

	glDisable(GL_DEPTH_TEST);

	FsGuiCanvas::Show();

	FsSwapBuffers();

	this->SetNeedRedraw(YSFALSE);
}



////////////////////////////////////////////////////////////



YSBOOL FsGui3dAppMustTerminate(void)
{
	return appMustTerminate;
}

int FsGui3dAppSleepIntervalInMS(void)
{
	return appSleepInterval;
}

static void FsGui3DAppLocalIntervalHook(void *)
{
	FsGui3DAppIntervalCallBack();
}

void FsGui3DAppInitialize(int argc,char *const argv[])
{
	appMainCanvas=new FsGui3DMainCanvas;
	YsGLSLCreateSharedRenderer();
	FsRegisterIntervalCallBack(FsGui3DAppLocalIntervalHook,NULL);

	if(2<=argc)
	{
		YsWString wStr;
		wStr.SetUTF8String(argv[1]);
		appMainCanvas->OpenFile(wStr);
	}
}

void FsGui3DAppIntervalCallBack(void)
{
	appMainCanvas->OnInterval();
}

void FsGui3dAppCleanUp(void)
{
	YsGLSLDeleteSharedRenderer();
	delete appMainCanvas;
}

