#ifndef YSTHREAD_IS_INCLUDED
#define YSTHREAD_IS_INCLUDED
/* { */

#include <ysclass.h>

class YsMutex
{
public:
	virtual ~YsMutex(){};

	virtual void Lock(void)=0;
	virtual void Unlock(void)=0;
	virtual YSRESULT LockNoWait(void)=0;
};


enum YSTHREADTASKSTATE
{
	YSTHREADTASKSTATE_IDLE,
	YSTHREADTASKSTATE_RUNNING,
	YSTHREADTASKSTATE_FINISHED
};

class YsThreadTask
{
friend class YsThreadController;

private:
	int taskId;

public:
	YSTHREADTASKSTATE sta;

	void EntryPoint(void);  // Actually called from YsThreadController
	int GetTaskId(void) const;
	virtual void Launch(void)=0;  // Implementation of the thread task.
};



class YsGenericThreadController
{
public:
	virtual YsMutex *CreateSingleMutex(void)=0;
	virtual YSRESULT DeleteSingleMutex(YsMutex *mutex)=0;

	virtual unsigned int GetNumCPU(void)=0;
	virtual void CacheThread(int nThread)=0;

	virtual YSRESULT RunParallel(int nTask,YsThreadTask *task[])=0;
};



class YsThreadController : public YsGenericThreadController
{
private:
	YsArray <class YsThreadCache *> threadCache;
public:
	YsThreadController();
	~YsThreadController();

	virtual YsMutex *CreateSingleMutex(void);
	virtual YSRESULT DeleteSingleMutex(YsMutex *mutex);

	virtual unsigned int GetNumCPU(void);
	virtual void CacheThread(int nThread);

	virtual YSRESULT RunParallel(int nTask,YsThreadTask *task[]);
};

/* } */
#endif
